# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy as copy
from string import ascii_lowercase as alphabet

import numpy as np
from scipy.special import factorial

from ._base import Element


def lagrange_line(order):
    "Return the cell-connectivity for an arbitrary-order Lagrange line."

    vertices = np.array([0, order])
    edge = np.arange(order + 1)[1:-1]

    return np.concatenate([vertices, edge])


def lagrange_quad(order):
    "Return the cell-connectivity for an arbitrary-order Lagrange quad."
    # points on a unit rectangle
    x = np.linspace(0, 1, order + 1)
    points = np.vstack([p.ravel() for p in np.meshgrid(x, x, indexing="ij")][::-1]).T

    # search vertices
    xmin = min(points[:, 0])
    ymin = min(points[:, 1])

    xmax = max(points[:, 0])
    ymax = max(points[:, 1])

    def search_vertice(p, x, y):
        return np.where(np.logical_and(p[:, 0] == x, p[:, 1] == y))[0][0]

    def search_edge(p, a, x):
        return np.where(p[:, a] == x)[0][1:-1]

    v1 = search_vertice(points, xmin, ymin)
    v2 = search_vertice(points, xmax, ymin)
    v3 = search_vertice(points, xmax, ymax)
    v4 = search_vertice(points, xmin, ymax)

    vertices = [v1, v2, v3, v4]

    mask1 = np.ones_like(points[:, 0], dtype=bool)
    mask1[vertices] = 0

    e1 = search_edge(points, 1, ymin)
    e2 = search_edge(points, 0, xmax)
    e3 = search_edge(points, 1, ymax)
    e4 = search_edge(points, 0, xmin)

    edges = np.hstack((e1, e2, e3, e4))

    mask2 = np.ones_like(points[:, 0], dtype=bool)
    mask2[np.hstack((vertices, edges))] = 0

    face = np.arange(len(points))[mask2]
    return np.hstack((vertices, edges, face))


def lagrange_hexahedron(order):
    "Return the cell-connectivity for an arbitrary-order Lagrange hexahedron."

    x = np.linspace(0, 1, order + 1)
    points = np.vstack([p.ravel() for p in np.meshgrid(x, x, x, indexing="ij")][::-1]).T

    # search vertices
    xmin = min(points[:, 0])
    ymin = min(points[:, 1])
    zmin = min(points[:, 2])

    xmax = max(points[:, 0])
    ymax = max(points[:, 1])
    zmax = max(points[:, 2])

    def search_vertice(p, x, y, z):
        return np.where(
            np.logical_and(np.logical_and(p[:, 0] == x, p[:, 1] == y), p[:, 2] == z)
        )[0][0]

    def search_edge(p, a, b, x, y):
        return np.where(np.logical_and(p[:, a] == x, p[:, b] == y))[0][1:-1]

    def search_face(p, a, x, vertices, edges):
        face = np.where(points[:, a] == x)[0]
        mask = np.zeros_like(p[:, 0], dtype=bool)
        mask[face] = 1
        mask[np.hstack((vertices, edges))] = 0
        return np.arange(len(p[:, 0]))[mask]

    v1 = search_vertice(points, xmin, ymin, zmin)
    v2 = search_vertice(points, xmax, ymin, zmin)
    v3 = search_vertice(points, xmax, ymax, zmin)
    v4 = search_vertice(points, xmin, ymax, zmin)

    v5 = search_vertice(points, xmin, ymin, zmax)
    v6 = search_vertice(points, xmax, ymin, zmax)
    v7 = search_vertice(points, xmax, ymax, zmax)
    v8 = search_vertice(points, xmin, ymax, zmax)

    vertices = [v1, v2, v3, v4, v5, v6, v7, v8]

    mask1 = np.ones_like(points[:, 0], dtype=bool)
    mask1[vertices] = 0

    e1 = search_edge(points, 1, 2, ymin, zmin)
    e2 = search_edge(points, 0, 2, xmax, zmin)
    e3 = search_edge(points, 1, 2, ymax, zmin)
    e4 = search_edge(points, 0, 2, xmin, zmin)

    e5 = search_edge(points, 1, 2, ymin, zmax)
    e6 = search_edge(points, 0, 2, xmax, zmax)
    e7 = search_edge(points, 1, 2, ymax, zmax)
    e8 = search_edge(points, 0, 2, xmin, zmax)

    e9 = search_edge(points, 0, 1, xmin, ymin)
    e10 = search_edge(points, 0, 1, xmax, ymin)
    e11 = search_edge(points, 0, 1, xmax, ymax)
    e12 = search_edge(points, 0, 1, xmin, ymax)

    edges = np.hstack((e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e11, e12))

    mask2 = np.ones_like(points[:, 0], dtype=bool)
    mask2[np.hstack((vertices, edges))] = 0

    f1 = search_face(points, 0, xmin, vertices, edges)
    f2 = search_face(points, 0, xmax, vertices, edges)
    f3 = search_face(points, 1, ymin, vertices, edges)
    f4 = search_face(points, 1, ymax, vertices, edges)
    f5 = search_face(points, 2, zmin, vertices, edges)
    f6 = search_face(points, 2, zmax, vertices, edges)

    faces = np.hstack((f1, f2, f3, f4, f5, f6))

    mask3 = np.ones_like(points[:, 0], dtype=bool)
    mask3[np.hstack((vertices, edges, faces))] = 0
    volume = np.arange(len(points))[mask3]

    return np.hstack((vertices, edges, faces, volume))


class ArbitraryOrderLagrange(Element):
    r"""A n-dimensional Lagrange finite element (e.g. line, quad or hexahdron) of
    arbitrary order.

    Notes
    -----

    **Polynomial shape functions**

    The basis function vector is generated with row-stacking of the individual lagrange
    polynomials. Each polynomial defined in the interval :math:`[-1,1]` is a function of
    the parameter :math:`r`. The curve parameters matrix :math:`\boldsymbol{A}` is of
    symmetric shape due to the fact that for each evaluation point :math:`r_j` exactly
    one basis function :math:`h_j(r)` is needed.

    ..  math::

        \boldsymbol{h}(r) = \boldsymbol{A}^T \boldsymbol{r}(r)

    **Curve parameter matrix**

    The evaluation of the curve parameter matrix :math:`\boldsymbol{A}` is carried out
    by boundary conditions. Each shape function :math:`h_i` has to take the value of one
    at the associated nodal coordinate :math:`r_i` and zero at all other point
    coordinates.

    ..  math::

        \boldsymbol{A}^T \boldsymbol{R} &=
            \boldsymbol{I} \qquad \text{with} \qquad \boldsymbol{R} =
            \begin{bmatrix}\boldsymbol{r}(r_1) & \boldsymbol{r}(r_2) & \dots &
            \boldsymbol{r}(r_p)\end{bmatrix}

        \boldsymbol{A}^T &= \boldsymbol{R}^{-1}


    **Interpolation and partial derivatives**

    The approximation of nodal unknowns :math:`\hat{\boldsymbol{u}}` as a function of
    the parameter :math:`r` is evaluated as

    ..  math::

        \boldsymbol{u}(r) \approx \hat{\boldsymbol{u}}^T \boldsymbol{h}(r)

    For the calculation of the partial derivative of the interpolation field w.r.t. the
    parameter :math:`r` a simple shift of the entries of the parameter vector is enough.
    This shifted parameter vector is denoted as :math:`\boldsymbol{r}^-`. A minus
    superscript indices the negative shift of the vector entries by :math:`-1`.

    ..  math::

        \frac{\partial \boldsymbol{u}(r)}{\partial r} &\approx
            \hat{\boldsymbol{u}}^T \frac{\partial \boldsymbol{h}(r)}{\partial r}

        \frac{\partial \boldsymbol{h}(r)}{\partial r} &=
            \boldsymbol{A}^T \boldsymbol{r}^-(r) \qquad \text{with} \qquad r_0^- =
            0 \qquad \text{and} \qquad r_i^- = \frac{r^{(i-1)}}{(i-1)!} \qquad
            \text{for} \qquad  i=(1 \dots p)


    n-dimensional shape functions
    *****************************

    Multi-dimensional shape function matrices
    :math:`\boldsymbol{H}_{2D}, \boldsymbol{H}_{3D}` are simply evaluated as dyadic
    (outer) vector products of one-dimensional shape function vectors. The multi-
    dimensional shape function vector is a one-dimensional representation (flattened
    version) of the multi-dimensional shape function matrix.

    ..  math::

        \boldsymbol{H}_{2D}(r,s) &= \boldsymbol{h}(r) \otimes \boldsymbol{h}(s)

        \boldsymbol{H}_{3D}(r,s,t) &= \boldsymbol{h}(r) \otimes \boldsymbol{h}(s)
            \otimes \boldsymbol{h}(t)

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> element = fem.ArbitraryOrderLagrangeElement(order=4, dim=2)
       >>> element.plot().show()

    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> element = fem.ArbitraryOrderLagrangeElement(order=3, dim=3)
       >>> element.plot().show()
    """

    def __init__(self, order, dim, interval=(-1, 1), permute=True):
        self._order = order
        self._nshape = order + 1
        self._npoints = self._nshape**dim
        self._nbasis = self._npoints
        self._interval = interval

        self.dim = dim

        self.permute = None
        if permute:
            self.permute = [None, lagrange_line, lagrange_quad, lagrange_hexahedron][
                dim
            ](order)

        # init curve-parameter matrix
        n = self._nshape
        self._AT = np.linalg.inv(
            np.array([self._polynomial(p, n) for p in self._points(n)])
        ).T

        # indices for outer product in einstein notation
        # idx = ["a", "b", "c", ...][:dim]
        # subscripts = "a,b,c -> abc"
        self._idx = [letter for letter in alphabet][:dim]
        self._subscripts = ",".join(self._idx) + "->" + "".join(self._idx)

        # init points
        grid = np.meshgrid(*np.tile(self._points(n), (dim, 1)), indexing="ij")[::-1]
        self.points = np.vstack([point.ravel() for point in grid]).T

        if self.permute is not None:
            self.points = self.points[self.permute]

        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "VTK_LAGRANGE"

        if dim == 1:
            self.cell_type += "_LINE"
        elif dim == 2:
            self.cell_type += "_QUADRILATERAL"
        elif dim == 3:
            self.cell_type += "_HEXAHEDRON"

    def function(self, r):
        "Return the shape functions at given coordinate vector r."
        n = self._nshape

        # 1d - basis function vectors per axis
        fun = [self._AT @ self._polynomial(ra, n) for ra in r]
        h = np.einsum(self._subscripts, *fun).ravel("F")

        if self.permute is not None:
            h = h[self.permute]

        return h

    def gradient(self, r):
        "Return the gradient of shape functions at given coordinate vector r."
        n = self._nshape

        # 1d - basis function vectors per axis
        h = [self._AT @ self._polynomial(ra, n) for ra in r]

        # shifted 1d - basis function vectors per axis
        k = [self._AT @ np.append(0, self._polynomial(ra, n)[:-1]) for ra in r]

        # init output
        dhdr = np.zeros((n**self.dim, self.dim))

        # loop over columns
        for i in range(self.dim):
            g = copy(h)
            g[i] = k[i]
            dhdr[:, i] = np.einsum(self._subscripts, *g).ravel("F")

        if self.permute is not None:
            dhdr = dhdr[self.permute]

        return dhdr

    def hessian(self, r):
        "Return the hessian of shape functions at given coordinate vector r."
        n = self._nshape

        # 1d - basis function vectors per axis
        h = [self._AT @ self._polynomial(ra, n) for ra in r]

        # shifted 1d - basis function vectors per axis
        h1 = [self._AT @ np.append(0, self._polynomial(ra, n)[:-1]) for ra in r]
        h2 = [self._AT @ np.append([0, 0], self._polynomial(ra, n)[:-2]) for ra in r]

        # init output
        d2hdrdr = np.zeros((n**self.dim, self.dim, self.dim))

        # loop over columns
        for i in range(self.dim):
            for j in range(self.dim):
                g = copy(h)
                if i == j:
                    g[i] = h2[i]
                else:
                    g[i] = h1[i]
                    g[j] = h1[j]
                d2hdrdr[:, i, j] = np.einsum(self._subscripts, *g).ravel("F")

        if self.permute is not None:
            d2hdrdr = d2hdrdr[self.permute]

        return d2hdrdr

    def _points(self, n):
        "Equidistant n-points in interval [-1, 1]."
        return np.linspace(*self._interval, n)

    def _polynomial(self, r, n):
        "Lagrange-Polynomial of order n evaluated at coordinate vector r."
        m = np.arange(n)
        return r**m / factorial(m)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from ._morph_uniaxial import morph_uniaxial
from .microsphere import affine_force_statevars


def morph_representative_directions(F, statevars, p, ε=1e-6):
    """First Piola-Kirchhoff stress tensor of the
    `MORPH <https://doi.org/10.1016/s0749-6419(02)00091-8>`_ model formulation [1]_,
    implemented by the concept of
    `representative directions <https://nbn-resolving.org/urn:nbn:de:bsz:ch1-qucosa-114428>`_
    [2]_, [3]_.

    Parameters
    ----------
    F : tensortrax.Tensor or jax.Array
        Deformation gradient tensor.
    statevars : array
        Vector of stacked state variables (CTS, λ - 1, SA1, SA2).
    p : list of float
        A list which contains the 8 material parameters.
    ε : float, optional
        A small stabilization parameter (default is 1e-6).

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> import felupe as fem
        >>>
        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Material(
        ...     mat.models.lagrange.morph_representative_directions,
        ...     p=[0.011, 0.408, 0.421, 6.85, 0.0056, 5.54, 5.84, 0.117],
        ...     nstatevars=84,
        ... )
        >>> ax = umat.plot(
        ...    incompressible=True,
        ...    ux=fem.math.linsteps(
        ...        # [1, 2, 1, 2.75, 1, 3.5, 1, 4.2, 1, 4.8, 1, 4.8, 1],
        ...        [1, 2.75, 1, 2.75],
        ...        num=20,
        ...    ),
        ...    ps=None,
        ...    bx=None,
        ... )

    References
    ----------
    .. [1] D. Besdo and J. Ihlemann, "A phenomenological constitutive model for
       rubberlike materials and its numerical applications", International Journal
       of Plasticity, vol. 19, no. 7. Elsevier BV, pp. 1019–1036, Jul. 2003. doi:
       `10.1016/s0749-6419(02)00091-8 <https://doi.org/10.1016/s0749-6419(02)00091-8>`_.

    .. [2] M. Freund, "Verallgemeinerung eindimensionaler Materialmodelle für die
       Finite-Elemente-Methode", Dissertation, Technische Universität Chemnitz,
       Chemnitz, 2013.

    .. [3] C. Miehe, S. Göktepe and F. Lulei, "A micro-macro approach to rubber-like
       materials - Part I: the non-affine micro-sphere model of rubber elasticity",
       Journal of the Mechanics and Physics of Solids, vol. 52, no. 11. Elsevier BV, pp.
       2617–2660, Nov. 2004. doi:
       `10.1016/j.jmps.2004.03.011 <https://doi.org/10.1016/j.jmps.2004.03.011>`_.

    See Also
    --------
    felupe.constitution.tensortrax.models.lagrange.morph : Strain energy function of the
        MORPH model formulation.
    felupe.constitution.jax.models.lagrange.morph : Strain energy function of the MORPH
        model formulation.
    """

    def f(λ, statevars, **kwargs):
        dψdλ, statevars_new = morph_uniaxial(λ, statevars, **kwargs)
        return 5 * dψdλ, statevars_new

    return affine_force_statevars(F, statevars, f=f, kwargs={"p": p, "ε": ε})



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from tensortrax.math import sum as tsum
from tensortrax.math import trace
from tensortrax.math.linalg import det


def lopez_pamies(C, mu, alpha):
    r"""Strain energy function of the isotropic hyperelastic
    `Lopez-Pamies <https://doi.org/10.1016/j.crme.2009.12.007>`_ material
    formulation [1]_.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    mu : list of float
        List of moduli.
    alpha : list of float
        List of invariant exponents.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-lp`

    ..  math::
        :label: psi-lp

        \psi = \sum_{r=1}^M \frac{3^{1-\alpha_r}}{2 \alpha_r} \mu_r \left(
            \hat{I}_1^{\alpha_r} - 3^{\alpha_r}
        \right)

    with the first main invariant of the distortional part of the right
    Cauchy-Green deformation tensor, see Eq. :eq:`invariant-lp`.

    ..  math::
        :label: invariant-lp

        \hat{I}_1 = J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

    The sum of the moduli :math:`\mu_r` is equal to the initial shear modulus
    :math:`\mu`, see Eq. :eq:`shear-modulus-lp`.

    ..  math::
        :label: shear-modulus-lp

        \mu = \sum_r \mu_r

    Examples
    --------

    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.Hyperelastic(
        ...     fem.lopez_pamies, mu=[0.2699,  0.00001771], alpha=[1.08, 4.40]
        ... )
        >>>
        >>> ux = fem.math.linsteps([0.6, 7], num=50)
        >>> ps = fem.math.linsteps([1, 7], num=50)
        >>> bx = fem.math.linsteps([1, 5], num=50)
        >>>
        >>> ax = umat.plot(ux=ux, ps=ps, bx=bx, incompressible=True)

    References
    ----------
    .. [1] O. Lopez-Pamies, "A new I1-based hyperelastic model for rubber elastic
       materials", Comptes Rendus. Mécanique, vol. 338, no. 1. Cellule MathDoc/Centre
       Mersenne, pp. 3–11, Dec. 23, 2009. doi:
       `10.1016/j.crme.2009.12.007 <https://doi.org/10.1016/j.crme.2009.12.007>`_.

    """
    I1 = det(C) ** (-1 / 3) * trace(C)
    ψr = lambda μr, αr, I1: 3 ** (1 - αr) / (2 * αr) * μr * (I1**αr - 3**αr)

    return tsum([ψr(μr, αr, I1) for μr, αr in zip(mu, alpha)])



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import matplotlib

matplotlib.use("Agg")

import matplotlib.pyplot as plt
import numpy as np
import pytest

import felupe as fem
import felupe.constitution.tensortrax.models as models


def pre(sym, add_identity, add_random=False):
    m = fem.mesh.Cube()
    e = fem.element.Hexahedron()
    q = fem.quadrature.GaussLegendre(1, 3)
    r = fem.Region(m, e, q)
    u = fem.Field(r, dim=3)

    if add_random:
        np.random.seed(55601)
        u.values += np.random.rand(*u.values.shape) / 20

    v = fem.FieldContainer([u])
    return r, v.extract(grad=True, sym=sym, add_identity=add_identity)


def pre_mixed(sym, add_identity):
    m = fem.mesh.Cube()
    e = fem.element.Hexahedron()
    q = fem.quadrature.GaussLegendre(1, 3)
    r = fem.Region(m, e, q)
    u = fem.Field(r, dim=3)
    v = fem.Field(r, dim=1)
    z = fem.Field(r, dim=1, values=1.0)
    w = fem.FieldContainer([u, v, z])
    return r, w.extract(grad=True, sym=sym, add_identity=add_identity), m


def test_nh():
    r, F = pre(sym=False, add_identity=True)

    for parallel in [False, True]:
        for nh in [
            fem.constitution.NeoHooke(mu=1.0, bulk=2.0, parallel=parallel),
            fem.constitution.NeoHookeCompressible(mu=1.0, lmbda=2.0, parallel=parallel),
            fem.constitution.LinearElasticLargeStrain(E=1.0, nu=0.3, parallel=parallel),
        ]:
            W = nh.function(F)
            P = nh.gradient(F)[:-1]
            A = nh.hessian(F)

            Wx = nh.energy(F)
            Px = nh.stress(F)[:-1]
            Ax = nh.elasticity(F)

            ax = nh.plot(incompressible=False)
            plt.close(ax.get_figure())

            assert np.allclose(W, Wx)
            assert np.allclose(P, Px)
            assert np.allclose(A, Ax)

            assert W[0].shape == F[0].shape[-2:]
            assert P[0].shape == (3, 3, *F[0].shape[-2:])
            assert A[0].shape == (3, 3, 3, 3, *F[0].shape[-2:])

            nh = fem.constitution.NeoHooke(mu=None, bulk=2.0, parallel=parallel)

            assert np.allclose(P, 0)

            nh = fem.constitution.NeoHooke(mu=1.0, parallel=parallel)

            W = nh.function(F)
            P = nh.gradient(F)[:-1]
            A = nh.hessian(F)

            assert W[0].shape == F[0].shape[-2:]
            assert P[0].shape == (3, 3, *F[0].shape[-2:])
            assert A[0].shape == (3, 3, 3, 3, *F[0].shape[-2:])

    assert np.all(nh.is_stable(F))


def test_linear():
    r, F = pre(sym=False, add_identity=True)

    check_stress = []
    check_dsde = []

    for Material in [
        (fem.constitution.LinearElastic, {}),
        (fem.constitution.LinearElasticTensorNotation, dict(parallel=False)),
        (fem.constitution.LinearElasticTensorNotation, dict(parallel=True)),
    ]:
        LinearElastic, kwargs = Material

        le = LinearElastic(E=1.0, nu=0.3, **kwargs)

        stress = le.gradient(F)[:-1]
        dsde = le.hessian(F)
        dsde2 = le.hessian(shape=F[0].shape[-2:])

        assert le.elasticity()[0].shape[-2:] == (1, 1)

        check_stress.append(stress)
        check_dsde.append([dsde[0][..., 0, 0], dsde2[0][..., 0, 0]])

        assert dsde[0].shape[:-2] == dsde2[0].shape[:-2]
        assert np.allclose(stress, 0)

    assert np.allclose(*check_stress)
    assert np.allclose(*check_dsde)


def test_linear_orthotropic():
    r, F = pre(sym=False, add_identity=True)

    lmbda, mu = fem.constitution.lame_converter_orthotropic(
        E=[6919, 271, 450],
        nu=[0.388, 0.278, 0.375],
        G=[262, 34, 354],
    )

    for Material in [
        (fem.constitution.LinearElasticOrthotropic, {}),
    ]:
        LinearElasticOrtho, kwargs = Material

        # doi.org/10.2478/ace-2018-0027 (pine wood)
        le = LinearElasticOrtho(
            E=[6919, 271, 450],
            nu=[0.388, 0.278, 0.375],
            G=[262, 34, 354],
            **kwargs,
        )

        stress = le.gradient(F)[:-1]
        dsde = le.hessian(F)

        assert le.elasticity()[0].shape[-2:] == (1, 1)

        assert stress[0].shape == (3, 3, *F[0].shape[-2:])
        assert dsde[0].shape == (3, 3, 3, 3, 1, 1)

        assert np.allclose(stress, 0)


def test_linear_planestress():
    r, F = pre(sym=False, add_identity=True)
    F = [F[0][:2][:, :2]]

    le = fem.constitutive_material(fem.constitution.LinearElasticPlaneStress)(
        E=1.0, nu=0.3
    )

    stress = le.gradient(F)[:-1]
    dsde = le.hessian(F)
    dsde = le.hessian(F)
    dsde2 = le.hessian(shape=F[0].shape[-2:])

    assert le.elasticity()[0].shape[-2:] == (1, 1)

    check_dsde = [dsde[0][..., 0, 0], dsde2[0][..., 0, 0]]

    assert np.allclose(*check_dsde)

    stress_full = le.stress(F)
    strain_full = le.strain(F)

    assert stress_full[0].shape == (3, 3, *F[0].shape[-2:])
    assert strain_full[0].shape == (3, 3, *F[0].shape[-2:])

    assert stress[0].shape == (2, 2, *F[0].shape[-2:])
    assert dsde[0].shape == (2, 2, 2, 2, 1, 1)

    assert np.allclose(stress, 0)


def test_linear_planestrain():
    r, F = pre(sym=False, add_identity=True)
    F = [F[0][:2][:, :2]]

    le = fem.constitution.LinearElasticPlaneStrain(E=1.0, nu=0.3)

    stress = le.gradient(F)[:-1]
    dsde = le.hessian(F)
    dsde = le.hessian(F)

    stress_full = le.stress(F)
    strain_full = le.strain(F)

    assert stress_full[0].shape == (3, 3, *F[0].shape[-2:])
    assert strain_full[0].shape == (3, 3, *F[0].shape[-2:])

    assert stress[0].shape == (2, 2, *F[0].shape[-2:])
    assert dsde[0].shape == (2, 2, 2, 2, 1, 1)

    assert np.allclose(stress, 0)


def test_kinematics():
    r, F = pre(sym=False, add_identity=True)

    N = F[0][:, 0]

    for parallel in [False, True]:
        lc = fem.constitution.LineChange(parallel=parallel)
        ac = fem.constitution.AreaChange(parallel=parallel)
        vc = fem.constitution.VolumeChange(parallel=parallel)

        xf = lc.function(F)
        xg = lc.gradient(F)
        xg = lc.gradient(F, parallel=parallel)

        Yf = ac.function(F, N)
        Yf = ac.function(F, N, parallel=parallel)
        Yg = ac.gradient(F, N)
        Yg = ac.gradient(F, N, parallel=parallel)

        yf = ac.function(F)
        yg = ac.gradient(F)

        zf = vc.function(F)
        zg = vc.gradient(F)
        zh = vc.hessian(F)
        zh = vc.hessian(F, parallel=parallel)

        assert np.allclose(xf, F)

        assert xf[0].shape == (3, 3, *F[0].shape[-2:])
        assert xg[0].shape == (3, 3, 3, 3, 1, 1)

        assert yf[0].shape == (3, 3, *F[0].shape[-2:])
        assert yg[0].shape == (3, 3, 3, 3, *F[0].shape[-2:])

        assert Yf[0].shape == (3, *F[0].shape[-2:])
        assert Yg[0].shape == (3, 3, 3, *F[0].shape[-2:])

        assert zf[0].shape == F[0].shape[-2:]
        assert zg[0].shape == (3, 3, *F[0].shape[-2:])
        assert zh[0].shape == (3, 3, 3, 3, *F[0].shape[-2:])


def test_umat():
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    from felupe.math import cdya, dya, identity, sym, trace

    def stress(x, mu, lmbda):
        "Evaluate the user-defined linear-elastic stress tensor."

        # extract variables
        F, statevars = x[0], x[-1]

        # user code for linear-elastic stress tensor
        e = sym(F)
        s = 2 * mu * e + lmbda * trace(e) * identity(e)

        # update state variables
        statevars_new = None

        return [s, statevars_new]

    def elasticity(x, mu, lmbda):
        """Evaluate the user-defined fourth-order elasticity tensor according to
        the first Piola-Kirchhoff stress tensor."""

        # extract variables
        F, statevars = x[0], x[-1]

        # user code for fourth-order elasticity tensor
        d = identity(F)
        dsde = 2 * mu * cdya(d, d) + lmbda * dya(d, d)

        return [dsde]

    linear_elastic = fem.Material(stress, elasticity, mu=1, lmbda=2)

    s, statevars_new = linear_elastic.gradient([F, None])
    dsde = linear_elastic.hessian([F, None])


def test_umat_hyperelastic_statevars(close_figs=True):
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    for model, kwargs, incompressible in [
        (
            fem.constitution.ogden_roxburgh,
            dict(r=3, m=1, beta=0, material=fem.neo_hooke, mu=1, nstatevars=1),
            True,
        ),
        (
            models.hyperelastic.morph_representative_directions,
            dict(
                p=[0.011, 0.408, 0.421, 6.85, 0.0056, 5.54, 5.84, 0.117], nstatevars=84
            ),
            True,
        ),
    ]:
        umat = fem.Hyperelastic(model, **kwargs)

        statevars = np.zeros((kwargs["nstatevars"], 8, 1))
        s, statevars_new = umat.gradient([F, statevars])
        dsde = umat.hessian([F, statevars])

        ux = fem.math.linsteps([1, 2, 1], num=10)
        ax = umat.plot(ux=ux, bx=None, ps=None, incompressible=True)

        if close_figs:
            plt.close(ax.get_figure())


def test_umat_hyperelastic(close_figs=True):
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    import tensortrax.math as tm

    def neo_hooke(C, mu=1):
        return mu / 2 * (tm.linalg.det(C) ** (-1 / 3) * tm.trace(C) - 3)

    for model, kwargs, incompressible in [
        (neo_hooke, {"mu": 1}, True),
        (fem.saint_venant_kirchhoff, {"mu": 1, "lmbda": 20.0}, False),
        (fem.saint_venant_kirchhoff, {"mu": 1, "lmbda": 20.0, "k": 0}, False),
        (fem.saint_venant_kirchhoff, {"mu": 1, "lmbda": 20.0, "k": 1}, False),
        (
            fem.saint_venant_kirchhoff_orthotropic,
            {
                "mu": [1, 1, 1],
                "lmbda": [20, 20, 20, 20, 20, 20],
                "r1": np.eye(3)[:, 0],
                "r2": np.eye(3)[:, 1],
                "r3": np.eye(3)[:, 2],
            },
            False,
        ),
        (
            fem.saint_venant_kirchhoff_orthotropic,
            {
                "mu": [1, 1, 1],
                "lmbda": [20, 20, 20, 20, 20, 20],
                "r1": np.eye(3)[:, 0],
                "r2": np.eye(3)[:, 1],
                "r3": None,
                "k": 0,
            },
            False,
        ),
        (
            fem.saint_venant_kirchhoff_orthotropic,
            {
                "mu": [1, 1, 1],
                "lmbda": [20, 20, 20, 20, 20, 20],
                "r1": np.eye(3)[:, 0],
                "r2": np.eye(3)[:, 1],
                "r3": np.eye(3)[:, 2],
                "k": 1,
            },
            False,
        ),
        (fem.neo_hooke, {"mu": 1}, True),
        (fem.mooney_rivlin, {"C10": 0.3, "C01": 0.8}, True),
        (fem.yeoh, {"C10": 0.5, "C20": -0.1, "C30": 0.02}, True),
        (
            fem.third_order_deformation,
            {"C10": 0.5, "C01": 0.1, "C11": 0.01, "C20": -0.1, "C30": 0.02},
            True,
        ),
        (fem.ogden, {"mu": [1, 0.2], "alpha": [1.7, -1.5]}, True),
        (fem.arruda_boyce, {"C1": 1.0, "limit": 3.2}, True),
        (
            fem.extended_tube,
            {"Gc": 0.1867, "Ge": 0.2169, "beta": 0.2, "delta": 0.09693},
            True,
        ),
        (
            fem.van_der_waals,
            {"mu": 1.0, "beta": 0.1, "a": 0.5, "limit": 5.0},
            True,
        ),
        (
            fem.alexander,
            dict(C1=0.117, C2=0.137, C3=0.00690, gamma=0.735, k=0.00015),
            True,
        ),
        (
            fem.anssari_benam_bucchi,
            dict(mu=0.29, N=26.8),
            True,
        ),
        (
            fem.lopez_pamies,
            dict(mu=[0.2699, 0.00001771], alpha=[1.08, 4.40]),
            True,
        ),
        (
            fem.storakers,
            dict(
                mu=[104.869e-3],
                alpha=[7.10874],
                beta=[0.106469],
            ),
            False,
        ),
        (fem.blatz_ko, dict(mu=1.0), False),
    ]:
        umat = fem.Hyperelastic(model, **kwargs)

        statevars = None
        s, statevars_new = umat.gradient([F, statevars])
        dsde = umat.hessian([F, statevars])

    for incompressible in [False, True]:
        ax = umat.plot(incompressible=incompressible)

        if close_figs:
            plt.close(ax.get_figure())

        ax = umat.screenshot(incompressible=incompressible)

        if close_figs:
            plt.close(ax.get_figure())

    umat = fem.Hyperelastic(fem.neo_hooke, mu=np.nan)

    with pytest.raises(ValueError):
        ax = umat.plot(bx=None, ps=None)

        if close_figs:
            plt.close(ax.get_figure())

    with pytest.raises(ValueError):
        ax = umat.plot(ux=None, bx=None)

        if close_figs:
            plt.close(ax.get_figure())

    with pytest.raises(ValueError):
        ax = umat.plot(ux=None, ps=None)

        if close_figs:
            plt.close(ax.get_figure())


def test_umat_hyperelastic2():
    r, x = pre(sym=False, add_identity=True, add_random=True)
    F = x[0]

    import tensortrax.math as tm

    def neo_hooke(F, mu=1):
        "First Piola-Kirchhoff stress of the Neo-Hookean material formulation."

        C = tm.dot(tm.transpose(F), F)
        Cu = tm.linalg.det(C) ** (-1 / 3) * C

        return mu * F @ tm.special.dev(Cu) @ tm.linalg.inv(C)

    kwargs = {"mu": 1}
    umat = fem.MaterialAD(neo_hooke, **kwargs)

    s, statevars_new = umat.gradient([F, None])
    dsde = umat.hessian([F, None])

    umat = fem.Hyperelastic(fem.constitution.neo_hooke, **kwargs)

    s2, statevars_new = umat.gradient([F, None])
    dsde2 = umat.hessian([F, None])

    assert np.allclose(s, s2)
    assert np.allclose(dsde, dsde2)


def test_umat_viscoelastic(close_figs=True):
    r, x = pre(sym=False, add_identity=True, add_random=True)
    F = x[0]

    import tensortrax.math as tm

    def viscoelastic(C, Cin, mu, eta, dtime):
        "Finite strain viscoelastic material formulation."
        Ci = (
            tm.special.from_triu_1d(Cin, like=C)
            + mu / eta * dtime * tm.linalg.det(C) ** (-1 / 3) * C
        )
        Ci = tm.linalg.det(Ci) ** (-1 / 3) * Ci
        I1 = tm.linalg.det(C) ** (-1 / 3) * tm.trace(C @ tm.linalg.inv(Ci))

        return mu / 2 * (I1 - 3), tm.special.triu_1d(Ci)

    kwargs = {"mu": 1, "eta": 1, "dtime": 1}
    umat = fem.Hyperelastic(viscoelastic, nstatevars=6, **kwargs)

    statevars = np.zeros((6, *F.shape[-2:]))
    s, statevars_new = umat.gradient([F, statevars])
    dsde = umat.hessian([F, statevars])

    umat = fem.Hyperelastic(
        fem.constitution.finite_strain_viscoelastic, nstatevars=6, **kwargs
    )
    ax = umat.plot(
        ux=fem.math.linsteps([1, 1.5, 1, 2, 1, 2.5, 1], num=15),
        ps=None,
        bx=None,
        incompressible=True,
    )

    if close_figs:
        plt.close(ax.get_figure())

    s2, statevars_new = umat.gradient([F, statevars])
    dsde2 = umat.hessian([F, statevars])

    assert np.allclose(s, s2)
    assert np.allclose(dsde, dsde2)


def test_umat_viscoelastic2():
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    import tensortrax.math as tm

    def viscoelastic(F, Cin, mu, eta, dtime):
        "Finite strain viscoelastic material formulation."
        C = tm.dot(tm.transpose(F), F)
        Cu = tm.linalg.det(C) ** (-1 / 3) * C
        Ci = (
            tm.special.from_triu_1d(Cin, like=C)
            + mu / eta * dtime * tm.linalg.det(C) ** (-1 / 3) * C
        )
        Ci = tm.linalg.det(Ci) ** (-1 / 3) * Ci
        S = mu * tm.special.dev(Cu @ tm.linalg.inv(Ci)) @ tm.linalg.inv(C)
        return F @ S, tm.special.triu_1d(Ci)

    kwargs = {"mu": 1, "eta": 1, "dtime": 1}
    umat = fem.MaterialAD(viscoelastic, nstatevars=6, **kwargs)

    statevars = np.zeros((6, *F.shape[-2:]))
    s, statevars_new = umat.gradient([F, statevars])
    dsde = umat.hessian([F, statevars])

    umat = fem.Hyperelastic(
        fem.constitution.finite_strain_viscoelastic, nstatevars=6, **kwargs
    )

    s2, statevars_new = umat.gradient([F, statevars])
    dsde2 = umat.hessian([F, statevars])

    assert np.allclose(s, s2)
    assert np.allclose(dsde, dsde2)


def test_umat_strain():
    r, x = pre(sym=False, add_identity=True)
    F = x[0]
    statevars = np.zeros((18 + 0, *F.shape[-2:]))

    for framework in ["small-strain", "total-lagrange", "co-rotational"]:

        umat = fem.MaterialStrain(
            material=fem.constitution.linear_elastic,
            λ=1,
            μ=2,
            framework=framework,
        )

        s, statevars_new = umat.gradient([F, statevars])
        dsde = umat.hessian([F, statevars])

    with pytest.raises(NotImplementedError):

        umat = fem.MaterialStrain(
            material=fem.constitution.linear_elastic,
            λ=1,
            μ=1,
            framework="my-fance-framework",
        )

        s, statevars_new = umat.gradient([F, statevars])
        dsde = umat.hessian([F, statevars])


def test_umat_strain_plasticity():
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    statevars = np.ones((18 + 10, *F.shape[-2:]))

    umat = fem.MaterialStrain(
        material=fem.constitution.linear_elastic_plastic_isotropic_hardening,
        λ=1,
        μ=1,
        σy=0,
        K=0.1,
        statevars=(1, (3, 3)),
    )

    s, statevars_new = umat.gradient([F, statevars])
    dsde = umat.hessian([F, statevars])


def test_umat_strain_viscoelasticity():
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    statevars = np.ones((18 + 18, *F.shape[-2:]))

    umat = fem.MaterialStrain(
        material=fem.constitution.linear_elastic_viscoelastic,
        λ=2.0,
        μ=1.0,
        G=[3.0, 15.0],
        τ=[10.0, 100.0],
        Δt=5.0,
        statevars=((2, 3, 3),),
    )

    s, statevars_new = umat.gradient([F, statevars])
    dsde = umat.hessian([F, statevars])


def test_elpliso():
    r, x = pre(sym=False, add_identity=True)
    F = x[0]
    statevars = np.zeros((28, *F.shape[-2:]))

    umat = fem.LinearElasticPlasticIsotropicHardening(E=3, nu=0.3, sy=1, K=0.1)

    s, statevars_new = umat.gradient([F, statevars])
    dsde = umat.hessian([F, statevars])


def test_composite(close_figs=True):
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    nh = fem.NeoHooke(mu=1.0)
    vol = fem.Volumetric(bulk=2.0)
    umat = nh & vol

    ax = umat.plot()

    if close_figs:
        plt.close(ax.get_figure())

    dWdF, statevars_new = umat.gradient([F, None])
    (d2WdFdF,) = umat.hessian([F, None])


def test_optimize(close_figs=True):
    stretches, stresses = (
        np.array(
            [
                [1.000, 0.00],
                [1.020, 0.26],
                [1.125, 1.37],
                [1.240, 2.30],
                [1.390, 3.23],
                [1.585, 4.16],
                [1.900, 5.10],
                [2.180, 6.00],
                [2.420, 6.90],
                [3.020, 8.80],
                [3.570, 10.7],
                [4.030, 12.5],
                [4.760, 16.2],
                [5.360, 19.9],
                [5.750, 23.6],
                [6.150, 27.4],
                [6.400, 31.0],
                [6.600, 34.8],
                [6.850, 38.5],
                [7.050, 42.1],
                [7.150, 45.8],
                [7.250, 49.6],
                [7.400, 53.3],
                [7.500, 57.0],
                [7.600, 64.4],
            ]
        )
        * np.array([1.0, 0.0980665])
    ).T

    for model in [
        fem.extended_tube,
    ]:
        umat = fem.Hyperelastic(model)
        with pytest.warns():
            umat_new, res = umat.optimize(ux=[stretches, stresses], incompressible=True)

        ux = np.linspace(stretches.min(), stretches.max(), num=200)
        ax = umat_new.plot(incompressible=True, ux=ux, bx=None, ps=None)
        ax.plot(stretches, stresses, "C0x")

        if close_figs:
            plt.close(ax.get_figure())

    for model in [
        fem.neo_hooke,
        fem.ogden,
        fem.third_order_deformation,
        fem.alexander,
    ]:
        umat = fem.Hyperelastic(model)
        umat_new, res = umat.optimize(ux=[stretches, stresses], incompressible=True)

        ux = np.linspace(stretches.min(), stretches.max(), num=200)
        ax = umat_new.plot(incompressible=True, ux=ux, bx=None, ps=None)
        ax.plot(stretches, stresses, "C0x")

        if close_figs:
            plt.close(ax.get_figure())

    for model in [
        fem.neo_hooke,
        fem.miehe_goektepe_lulei,
    ]:
        umat = fem.Hyperelastic(model)
        umat_new, res = umat.optimize(
            ux=[stretches, stresses], incompressible=True, relative=True
        )

        ux = np.linspace(stretches.min(), stretches.max(), num=200)
        ax = umat_new.plot(incompressible=True, ux=ux, bx=None, ps=None)
        ax.plot(stretches, stresses, "C0x")

        if close_figs:
            plt.close(ax.get_figure())


def test_lagrange():
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    import tensortrax.math as tm

    @fem.total_lagrange
    def neo_hooke_total_lagrange(F, mu=1):
        C = F.T @ F
        S = mu * tm.special.dev(tm.linalg.det(C) ** (-1 / 3) * C) @ tm.linalg.inv(C)
        return S

    @fem.updated_lagrange
    def neo_hooke_updated_lagrange(F, mu=1):
        b = F @ F.T
        J = tm.linalg.det(F)
        σ = mu * tm.special.dev(J ** (-2 / 3) * b) / J
        return σ

    for fun in [neo_hooke_total_lagrange, neo_hooke_updated_lagrange]:
        umat = fem.MaterialAD(fun, mu=1)
        nh = fem.NeoHooke(mu=1)

        P = umat.gradient([F])
        A4 = umat.hessian([F])
        assert np.allclose(P[0], nh.gradient([F])[0])
        assert np.allclose(A4[0], nh.hessian([F])[0])


def test_lagrange_statevars():
    r, x = pre(sym=False, add_identity=True)
    F = x[0]

    import tensortrax.math as tm

    @fem.total_lagrange
    def neo_hooke_total_lagrange(F, statevars, mu=1):
        C = F.T @ F
        S = mu * tm.special.dev(tm.linalg.det(C) ** (-1 / 3) * C) @ tm.linalg.inv(C)
        return S, statevars

    @fem.updated_lagrange
    def neo_hooke_updated_lagrange(F, statevars, mu=1):
        b = F @ F.T
        J = tm.linalg.det(F)
        σ = mu * tm.special.dev(J ** (-2 / 3) * b) / J
        return σ, statevars

    for fun in [neo_hooke_total_lagrange, neo_hooke_updated_lagrange]:
        umat = fem.MaterialAD(fun, mu=1, nstatevars=1)
        nh = fem.NeoHooke(mu=1)

        statevars = np.zeros((1, 8, 1))
        P = umat.gradient([F, statevars])
        A4 = umat.hessian([F, statevars])
        assert np.allclose(P[0], nh.gradient([F])[0])
        assert np.allclose(A4[0], nh.hessian([F])[0])

    p = [0.039, 0.371, 0.174, 2.41, 0.0094, 6.84, 5.65, 0.244]
    umat = fem.MaterialAD(fem.morph, p=p, nstatevars=13)

    statevars = np.zeros((13, 8, 1))
    P = umat.gradient([F, statevars])
    A4 = umat.hessian([F, statevars])

    assert not np.isnan(P[0]).any()
    assert not np.isnan(A4[0]).any()

    p = [0.011, 0.408, 0.421, 6.85, 0.0056, 5.54, 5.84, 0.117]
    umat = fem.MaterialAD(fem.morph_representative_directions, p=p, nstatevars=84)

    statevars = np.zeros((84, 8, 1))
    P = umat.gradient([F, statevars])
    A4 = umat.hessian([F, statevars])

    assert not np.isnan(P[0]).any()
    assert not np.isnan(A4[0]).any()


def test_laplace():
    r, F = pre(sym=False, add_identity=True)

    umat = fem.Laplace()
    W = umat.function(F)
    P = umat.gradient(F)[:-1]
    A = umat.hessian(F)

    assert W[0].shape == F[0].shape[-2:]
    assert P[0].shape == (3, 3, *F[0].shape[-2:])
    assert A[0].shape == (3, 3, 3, 3, 1, 1)


def test_plot_negative_stretches(close_figs=True):
    stretches = np.linspace(-0.5, 1, 16)
    umat = fem.NeoHooke(mu=1.0, bulk=2.0)

    for incompressible in [False, True]:
        with pytest.raises(ValueError):
            ax = umat.plot(ux=stretches, incompressible=incompressible)

            if close_figs:
                plt.close(ax.get_figure())

        with pytest.raises(ValueError):
            ax = umat.plot(bx=stretches, incompressible=incompressible)

            if close_figs:
                plt.close(ax.get_figure())

        with pytest.raises(ValueError):
            ax = umat.plot(ps=stretches, incompressible=incompressible)

            if close_figs:
                plt.close(ax.get_figure())


if __name__ == "__main__":

    close_figs = True

    test_nh()
    test_linear()
    test_linear_orthotropic()
    test_linear_planestress()
    test_linear_planestrain()
    test_kinematics()
    test_umat()
    test_umat_hyperelastic(close_figs=close_figs)
    test_umat_hyperelastic2()
    test_umat_hyperelastic_statevars(close_figs=close_figs)
    test_umat_viscoelastic(close_figs=close_figs)
    test_umat_viscoelastic2()
    test_umat_strain()
    test_umat_strain_plasticity()
    test_elpliso()
    test_composite(close_figs=close_figs)
    test_optimize(close_figs=close_figs)
    test_lagrange()
    test_lagrange_statevars()
    test_laplace()
    test_plot_negative_stretches(close_figs=True)



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import os

import numpy as np
import pytest

import felupe as fem


def pre():
    mesh = fem.Rectangle(n=2)
    region = fem.RegionQuad(mesh)
    field = fem.FieldsMixed(region, n=3, axisymmetric=True)

    umat = fem.ThreeFieldVariation(fem.NeoHooke(1, 5000))
    body = fem.SolidBody(umat, field)
    boundaries = fem.dof.uniaxial(field, return_loadcase=False)

    points = mesh.points[:, 0] == 1
    load = fem.PointLoad(field, points)
    gravity = fem.SolidBodyForce(field, [0, 0, 0], 0)

    region2 = fem.RegionQuadBoundary(mesh, mask=points, ensure_3d=True)
    field2 = fem.FieldContainer([fem.FieldAxisymmetric(region2, dim=2)])
    pressure = fem.SolidBodyPressure(field2, pressure=0.0)

    step = fem.Step(
        items=[body, load, gravity, pressure],
        ramp={
            boundaries["move"]: fem.math.linsteps([0, 1], num=10),
            load: np.zeros((11, 2)),
            pressure: np.zeros(11),
            gravity: np.zeros((11, 3)),
        },
        boundaries=boundaries,
    )

    return field, step


def weather(i, j, res, outside):
    assert outside == "rainy"


def test_job():
    field, step = pre()
    job = fem.Job(steps=[step])
    job.evaluate()
    field, step = pre()
    job = fem.Job(steps=[step], callback=weather, outside="rainy")
    job.evaluate(
        parallel=True,
        kwargs={"parallel": False},
        verbose=0,
    )


def test_job_xdmf():
    field, step = pre()

    job = fem.Job(steps=[step])
    job.evaluate()

    field, step = pre()
    job = fem.Job(steps=[step])
    job.evaluate(filename="result.xdmf", parallel=True, verbose=2)


def test_job_xdmf_global_field():
    field, step = pre()
    job = fem.Job(steps=[step])
    job.evaluate()

    field, step = pre()
    job = fem.Job(steps=[step])
    job.evaluate(filename="result.xdmf", x0=field, tqdm="auto")


def test_job_xdmf_vertex():

    import felupe as fem

    meshes = [
        fem.Cube(n=3),
        fem.Cube(n=3).translate(1, axis=0),
    ]
    container = fem.MeshContainer(meshes, merge=True)
    field = fem.Field.from_mesh_container(container).as_container()

    regions = [
        fem.RegionHexahedron(container.meshes[0]),
        fem.RegionHexahedron(container.meshes[1]),
    ]
    fields = [
        fem.FieldContainer([fem.Field(regions[0], dim=3)]),
        fem.FieldContainer([fem.Field(regions[1], dim=3)]),
    ]

    boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)
    umat = fem.LinearElasticLargeStrain(E=2.1e5, nu=0.3)
    solids = [
        fem.SolidBody(umat=umat, field=fields[0]),
        fem.SolidBody(umat=umat, field=fields[1]),
    ]

    move = fem.math.linsteps([0, 1], num=5)
    ramp = {boundaries["move"]: move}
    step = fem.Step(items=solids, ramp=ramp, boundaries=boundaries)

    job = fem.Job(steps=[step])

    with pytest.warns(UserWarning):
        job.evaluate(x0=field, filename="result.xdmf")


def test_curve():
    field, step = pre()

    curve = fem.CharacteristicCurve(
        steps=[step],
        boundary=step.boundaries["move"],
        callback=weather,
        outside="rainy",
    )

    with pytest.raises(ValueError):
        curve.plot()

    os.environ["FELUPE_VERBOSE"] = "true"

    curve.evaluate()
    curve.plot(xaxis=0, yaxis=0)
    curve.plot(x=np.zeros((10, 2)), y=np.ones((10, 2)), xaxis=0, yaxis=0)

    stretch = 1 + np.array(curve.x)[:, 0]
    area = 1**2 * np.pi
    force = (stretch - 1 / stretch**2) * area

    os.environ.pop("FELUPE_VERBOSE")

    assert np.allclose(np.array(curve.y)[:, 0], force, rtol=0.01)


def test_curve2():
    field, step = pre()

    curve = fem.CharacteristicCurve(steps=[step], boundary=step.boundaries["move"])
    curve.evaluate()
    curve.plot(xaxis=0, yaxis=0)

    stretch = 1 + np.array(curve.x)[:, 0]
    area = 1**2 * np.pi
    force = (stretch - 1 / stretch**2) * area

    assert np.allclose(np.array(curve.y)[:, 0], force, rtol=0.01)


def test_curve_custom_items():
    field, step = pre()

    curve = fem.CharacteristicCurve(
        steps=[step], items=step.items, boundary=step.boundaries["move"]
    )
    curve.evaluate()
    fig, ax = curve.plot(
        xaxis=0, yaxis=0, gradient=True, swapaxes=True, xlabel="x", ylabel="y"
    )
    curve.plot(x=np.zeros((10, 2)), y=np.ones((10, 2)), xaxis=0, yaxis=0, ax=ax)

    stretch = 1 + np.array(curve.x)[:, 0]
    area = 1**2 * np.pi
    force = (stretch - 1 / stretch**2) * area

    assert np.allclose(np.array(curve.y)[:, 0], force, rtol=0.01)


def test_empty():
    mesh = fem.Cube(n=2)
    region = fem.RegionHexahedron(mesh)
    field = fem.FieldsMixed(region, n=1)

    umat = fem.NeoHooke(mu=1, bulk=5000)
    solid = fem.SolidBody(umat, field)

    step = fem.Step(items=[solid], ramp=None, boundaries=None)
    job = fem.Job(steps=[step])

    with pytest.raises(ValueError):
        job.evaluate(tqdm="my_fancy_backend")

    job.evaluate(tqdm="notebook")


def test_noramp():
    mesh = fem.Cube(n=2)
    region = fem.RegionHexahedron(mesh)
    field = fem.FieldsMixed(region, n=1)

    umat = fem.LinearElastic(E=1, nu=0.3)
    solid = fem.SolidBody(umat, field)
    bounds = fem.dof.uniaxial(field, return_loadcase=False)

    step = fem.Step(items=[solid], ramp=None, boundaries=bounds)
    job = fem.Job(steps=[step])
    job.evaluate()


if __name__ == "__main__":
    test_job()
    test_job_xdmf()
    test_job_xdmf_global_field()
    test_job_xdmf_vertex()
    test_curve()
    test_curve2()
    test_curve_custom_items()
    test_empty()
    test_noramp()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ..math import right_cauchy_green_deformation as right_cauchy_green
from ..math import strain, strain_stretch_1d


class EvaluateFieldContainer:
    r"""Methods to evaluate the deformation gradient and strain measures of a field
    container.

    Parameters
    ----------
    field : FieldContainer
        A container for fields.

    Examples
    --------
    ..  pyvista-plot::

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle(n=4)
        >>> region = fem.RegionQuad(mesh)
        >>> field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])
        >>>
        >>> evaluate = fem.field.EvaluateFieldContainer(field)
        >>> F = evaluate.deformation_gradient()
        >>>
        >>> F.shape  # (3, 3, nquadraturepoints, ncells)
        (3, 3, 4, 9)

        >>> F[..., 0, 0]  # deformation gradient of first cell, first quadrature point
        array([[1., 0., 0.],
               [0., 1., 0.],
               [0., 0., 1.]])

    See Also
    --------
    felupe.FieldContainer : A container which holds one or multiple (mixed) fields.
    """

    def __init__(self, field):
        self.field = field

    def deformation_gradient(self):
        r"""Return the Deformation gradient tensor.

        .. math::
           :label: deformation-gradient-tensor

           \boldsymbol{F} &= \frac{\partial \boldsymbol{x}}{\partial \boldsymbol{X}}

           \boldsymbol{F} &= \sum_\alpha \lambda_\alpha
               \ \boldsymbol{n}_\alpha \otimes \boldsymbol{N}_\alpha

        """
        return self.field[0].extract()

    def right_cauchy_green_deformation(self):
        r"""Return the right Cauchy-Green deformation tensor.

        .. math::
           :label:right-cauchy-green-deformation-tensor

           \boldsymbol{F} &= \frac{\partial \boldsymbol{x}}{\partial \boldsymbol{X}}

           \boldsymbol{C} &= \boldsymbol{F}^T \boldsymbol{F}

           \boldsymbol{C} &= \sum_\alpha \lambda^2_\alpha
               \ \boldsymbol{N}_\alpha \otimes \boldsymbol{N}_\alpha

        """
        return right_cauchy_green(self.field)

    def strain(self, fun=strain_stretch_1d, tensor=True, asvoigt=False, n=0, **kwargs):
        r"""Return the Lagrangian strain tensor or its principal values.

        .. math::
           :label: lagrangian-strain-tensor

           \boldsymbol{E} = \sum_\alpha f_\alpha \left( \lambda_\alpha \right)
               \ \boldsymbol{N}_\alpha \otimes \boldsymbol{N}_\alpha

        By default, the Seth-Hill strain-stretch relation with a strain exponent of
        zero is used, see Eq. :eq:`seth-hill-strain-tensor`.

        .. math::
           :label: seth-hill-strain-tensor

           \boldsymbol{E} = \sum_\alpha \frac{1}{k} \left( \lambda_\alpha^k - 1 \right)
               \ \boldsymbol{N}_\alpha \otimes \boldsymbol{N}_\alpha

        Parameters
        ----------
        fun : callable, optional
            A callable for the one-dimensional strain-stretch relation. Its Signature
            must be ``lambda stretch, **kwargs: strain`` (default is the log. strain,
            :func:`~felupe.math.strain_stretch_1d` with ``k=0``).
        tensor : bool, optional
            Assemble and return the strain tensor if True or return its principal values
            only if False. Default is True.
        asvoigt : bool, optional
            Return the symmetric strain tensor in reduced vector storage (default is
            False).
        n : int, optional
            The index of the displacement field (default is 0).
        **kwargs : dict, optional
            Optional keyword-arguments are passed to the 1d strain-stretch relation.

        Returns
        -------
        ndarray of shape (N, N, ...) tensor, (N!, ...) asvoigt or (N, ...) princ. values
            The strain tensor or its principal values.

        See Also
        -------
        math.strain : Compute a Lagrangian strain tensor.
        math.strain_stretch_1d : Compute the Seth-Hill strains.
        """
        return strain(
            self.field, fun=fun, tensor=tensor, asvoigt=asvoigt, n=0, **kwargs
        )

    def log_strain(self, tensor=True, asvoigt=False, n=0):
        r"""Return the logarithmic Lagrangian strain tensor or its principal values.
        
        .. math::
           :label: log-strain-tensor

           \boldsymbol{E} = \sum_\alpha \ln(\lambda_\alpha) \
               \boldsymbol{N}_\alpha \otimes \boldsymbol{N}_\alpha

        Parameters
        ----------
        tensor : bool, optional
            Assemble and return the strain tensor if True or return its principal values
            only if False. Default is True.
        asvoigt : bool, optional
            Return the symmetric strain tensor in reduced vector storage (default is
            False).
        n : int, optional
            The index of the displacement field (default is 0).

        Returns
        -------
        ndarray of shape (N, N, ...) tensor, (N!, ...) asvoigt or (N, ...) princ. values
            The strain tensor or its principal values.

        See Also
        -------
        math.strain_stretch_1d : Compute the Seth-Hill strains.
        math.strain : Compute a Lagrangian strain tensor.
        """
        return strain(self.field, tensor=tensor, asvoigt=asvoigt, k=0)

    def green_lagrange_strain(self, tensor=True, asvoigt=False, n=0):
        r"""Return the Green-Lagrange Lagrangian strain tensor or its principal values.

        .. math::
           :label: seth-hill-strain-tensor

           \boldsymbol{E} = \sum_\alpha \frac{1}{2} \left( \lambda_\alpha^2 - 1 \right)
               \ \boldsymbol{N}_\alpha \otimes \boldsymbol{N}_\alpha

        Parameters
        ----------
        tensor : bool, optional
            Assemble and return the strain tensor if True or return its principal values
            only if False. Default is True.
        asvoigt : bool, optional
            Return the symmetric strain tensor in reduced vector storage (default is
            False).
        n : int, optional
            The index of the displacement field (default is 0).

        Returns
        -------
        ndarray of shape (N, N, ...) tensor, (N * (N + 1) / 2, ...) asvoigt or (N, ...) princ. values
            The strain tensor or its principal values.

        See Also
        -------
        math.strain : Compute a Lagrangian strain tensor.
        math.strain_stretch_1d : Compute the Seth-Hill strains.
        """
        return strain(self.field, tensor=tensor, asvoigt=asvoigt, k=2)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from functools import wraps

import numpy as np

from ..view import ViewMesh
from ._convert import (
    add_midpoints_edges,
    add_midpoints_faces,
    add_midpoints_volumes,
    cell_types,
    collect_edges,
    collect_faces,
    collect_volumes,
    convert,
    subdivide,
)
from ._discrete_geometry import DiscreteGeometry
from ._dual import dual
from ._tools import (
    expand,
    fill_between,
    flip,
    merge_duplicate_cells,
    merge_duplicate_points,
    mirror,
    revolve,
    rotate,
    runouts,
    translate,
    triangulate,
)


def as_mesh(obj):
    "Convert a ``DiscreteGeometry`` object to a ``Mesh`` object."
    return Mesh(points=obj.points, cells=obj.cells, cell_type=obj.cell_type)


class Mesh(DiscreteGeometry):
    r"""A mesh with points, cells and optional a specified cell type.

    Parameters
    ----------
    points : ndarray
        Array with point coordinates of shape ``(npoints, dim)``.
    cells : ndarray
        Array with cell connectivities of shape ``(ncells, points_per_cell)``.
    cell_type : str or None, optional
        An optional string in VTK-convention that specifies the cell type (default is
        None). Necessary when a mesh is saved to a file.

    Notes
    -----
    The ``points``-array is of shape ``(npoints, dim)``, as denoted in Eq. :eq:`points`.

    ..  math::
        :label: points

        \hat{\boldsymbol{X}} = \begin{bmatrix}
            \hat{\boldsymbol{X}_1}^T \\
            \hat{\boldsymbol{X}_2}^T \\
            \dots \\
            \hat{\boldsymbol{X}_n}^T \\
        \end{bmatrix}

    Exemplaric point coordinates of a single point at the origin in :math:`\mathbb{R}^3`
    are shown in Eq. :eq:`point`.

    ..  math::
        :label: point

        \hat{\boldsymbol{X}_1}^T = \begin{bmatrix} 0.0 & 0.0 & 0.0 \end{bmatrix}

    Topologic cell connectivies are stored inside the ``cells``-array, see Eq.
    :eq:`cells`,

    ..  math::
        :label: cells

        \boldsymbol{t} = \begin{bmatrix}
            \boldsymbol{t}_1^T \\
            \boldsymbol{t}_2^T \\
            \vdots \\
            \boldsymbol{t}_n^T \\
        \end{bmatrix}

    with an exemplaric topologic cell-connectivity of a :class:`~felupe.Quad`, as shown
    in Eq. :eq:`cell`.

    ..  math::
        :label: cell

        \boldsymbol{t}_1^T = \begin{bmatrix} 0 & 1 & 2 & 3 \end{bmatrix}

    Examples
    --------
    .. pyvista-plot::
       :context:
       :force_static:

       >>> import numpy as np
       >>> import felupe as fem
       >>>
       >>> points = np.array(
       ...     [[0.0, 0.0], [0.5, 0.1], [1.0, 0.2], [0.0, 1.0], [0.5, 0.9], [1.0, 0.8]]
       ... )
       >>> cells = np.array([[0, 1, 4, 3], [1, 2, 5, 4]])
       >>> mesh = fem.Mesh(points, cells, cell_type="quad")
       >>> mesh.plot().show()

       A mesh provides several useful attributes, like the dimension, the number of
       points, number of cells, number of deegrees of freedom, points connected to cells
       and points not connected to cells.

    .. pyvista-plot::
       :context:
       :force_static:

       >>> mesh.dim
       2

       >>> mesh.npoints
       6

       >>> mesh.ncells
       2

       >>> mesh.ndof
       12

       >>> mesh.points_with_cells
       array([0, 1, 2, 3, 4, 5])

    See Also
    --------
    felupe.MeshContainer : A container which operates on a list of meshes with identical
        dimensions.
    felupe.Rectangle : A rectangular 2d-mesh with quads between ``a`` and ``b`` with
        ``n`` points per axis.
    felupe.Cube : A cube shaped 3d-mesh with hexahedrons between ``a`` and ``b`` with
        ``n`` points per axis.
    felupe.Grid : A grid shaped 3d-mesh with hexahedrons. Basically a wrapper for
        :func:`numpy.meshgrid` with  default ``indexing="ij"``.
    felupe.Circle : A circular shaped 2d-mesh with quads and ``n`` points on the
        circumferential edge of a 45-degree section. 90-degree ``sections`` are placed
        at given angles in degree.
    felupe.mesh.Triangle : A triangular shaped 2d-mesh with quads and ``n`` points at
        the edges of the three sub-quadrilaterals.

    """

    def __init__(self, points, cells, cell_type=None):
        self.points = np.array(points)
        self.cells = np.array(cells)
        self.cell_type = cell_type

        super().__init__(points=points, cells=cells, cell_type=cell_type)

        self.__mesh__ = Mesh

        # alias
        self.sweep = self.merge_duplicate_points
        self.save = self.write
        self.as_pyvista = self.as_unstructured_grid

    def __repr__(self):
        header = "<felupe Mesh object>"
        points = f"  Number of points: {len(self.points)}"
        cells_header = "  Number of cells:"
        cells = [f"    {self.cell_type}: {self.ncells}"]

        return "\n".join([header, points, cells_header, *cells])

    def __str__(self):
        return self.__repr__()

    def disconnect(self, points_per_cell=None, calc_points=True):
        """Return a new instance of a Mesh with disconnected cells. Optionally, the
        points-per-cell may be specified (must be lower or equal the number of points-
        per-cell of the original Mesh). If the Mesh is to be used as a *dual* Mesh, then
        the point-coordinates do not have to be re-created because they are not used.

        See Also
        --------
        felupe.Mesh.dual : Create a new dual mesh with given points per cell.
        """

        return self.dual(
            points_per_cell=points_per_cell,
            disconnect=True,
            calc_points=calc_points,
            offset=0,
            npoints=None,
        )

    def as_meshio(self, **kwargs):
        """Export the mesh as :class:`meshio.Mesh`.

        Parameters
        ----------
        **kwargs : dict, optional
            Additional keyword-arguments for ``meshio.Mesh(points, cells, **kwargs)``.

        Returns
        -------
        meshio.Mesh
            The mesh as :class:`meshio.Mesh`.
        """

        import meshio

        points = np.pad(self.points, ((0, 0), (0, 3 - self.points.shape[1])))

        return meshio.Mesh(points=points, cells={self.cell_type: self.cells}, **kwargs)

    def as_unstructured_grid(self, cell_type=None, **kwargs):
        """Export the mesh as :class:`pyvista.UnstructuredGrid`.

        Parameters
        ----------
        cell_type : pyvista.CellType or None, optional
            Cell-type of PyVista (default is None).
        **kwargs : dict, optional
            Additional keyword-arguments for :class:`pyvista.UnstructuredGrid`.

        Returns
        -------
        pyvista.UnstructuredGrid
            The mesh as :class:`pyvista.UnstructuredGrid`.
        """

        import pyvista as pv

        if cell_type is None:
            cell_type = dict(cell_types())[self.cell_type]

        points = np.pad(self.points, ((0, 0), (0, 3 - self.points.shape[1])))
        cells = np.pad(
            self.cells, ((0, 0), (1, 0)), constant_values=self.cells.shape[1]
        )
        cell_types_int = cell_type * np.ones(self.ncells, dtype=int)

        return pv.UnstructuredGrid(cells, cell_types_int, points)

    def write(self, filename="mesh.vtk", **kwargs):
        """Write the mesh to a file.

        Parameters
        ----------
        filename : str, optional
            The filename of the mesh (default is ``mesh.vtk``).
        **kwargs : dict, optional
            Additional keyword arguments for :meth:`meshio.Mesh.write`.

        Notes
        -----
        ..  note::
            For XDMF-export please ensure to have ``h5py`` (as an optional dependency of
            ``meshio``) installed.

        Examples
        --------
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle(n=3)
        >>> mesh.write(filename="mesh.vtk")

        See Also
        --------
        felupe.mesh.read : Read a mesh from a file using :func:`meshio.read`.
        felupe.Mesh.write : Write the mesh to a file.

        """

        self.as_meshio(**kwargs).write(filename)

    def view(self, point_data=None, cell_data=None, cell_type=None):
        """View the mesh with optional given dicts of point- and cell-data items.

        Parameters
        ----------
        point_data : dict or None, optional
            Additional point-data dict (default is None).
        cell_data : dict or None, optional
            Additional cell-data dict (default is None).
        cell_type : pyvista.CellType or None, optional
            Cell-type of PyVista (default is None).

        Returns
        -------
        felupe.ViewMesh
            A object which provides visualization methods for :class:`felupe.Mesh`.

        See Also
        --------
        felupe.ViewMesh : Visualization methods for :class:`~felupe.Mesh`.
        """

        return ViewMesh(
            self,
            point_data=point_data,
            cell_data=cell_data,
            cell_type=cell_type,
        )

    def plot(self, *args, **kwargs):
        """Plot the mesh.

        See Also
        --------
        felupe.Scene.plot : Plot method of a scene.
        """
        return self.view().plot(*args, show_undeformed=False, **kwargs)

    def screenshot(
        self,
        *args,
        filename="mesh.png",
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the mesh.

        See Also
        --------
        pyvista.Plotter.screenshot : Take a screenshot of a PyVista plotter.
        """

        return self.plot(*args, off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )

    def imshow(self, *args, ax=None, **kwargs):
        """Take a screenshot of the mesh, show the image data in a figure and return the
        ax.
        """

        if ax is None:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()

        ax.imshow(self.screenshot(*args, filename=None, **kwargs))
        ax.set_axis_off()

        return ax

    def add_points(self, points):
        "Add additional points to the mesh in-place."
        self.update(points=np.vstack([self.points, points]))

    def clear_points_without_cells(self):
        "Clear the list of points without cells."
        self.points_without_cells = self.points_without_cells[:0]

    def get_point_ids(self, value, fun=np.isclose, mode=np.all, **kwargs):
        """Return point ids for points which are close to a given value.

        Parameters
        ----------
        value : float, list or ndarray
            Scalar value or point coordinates.
        fun : callable, optional
            The function used to compare the points to the given value, i.e. with a
            function signature ``fun(mesh.points, value, **kwargs)``. Default is
            :func:`numpy.isclose`.
        mode : callable, optional
            A callable used to combine the search results, either :func:`numpy.any` or
            :func:`numpy.all`.
        **kwargs : dict, optional
            Additional keyword arguments for ``fun(mesh.points, value, **kwargs)``.

        Returns
        -------
        ndarray
            Array with point ids.

        Examples
        --------
        Get point ids at given coordinates for a mesh with duplicate points.

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=11)
        >>> mesh.update(points=np.vstack([mesh.points, mesh.points]))
        >>> point_ids = mesh.get_point_ids([0, 1, 1])
        >>> point_ids
        array([1320, 2651])

        >>> mesh.points[point_ids]
        array([[0., 1., 1.],
               [0., 1., 1.]])

        """
        return np.argwhere(mode(fun(self.points, value, **kwargs), axis=1))[:, 0]

    def get_cell_ids(self, point_ids):
        """Return cell ids which have the given point ids in their connectivity.

        Parameters
        ----------
        point_ids : list or ndarray
            Array with point ids which are used to search for cells.

        Returns
        -------
        ndarray
            Array with cell ids which have the given point ids in their connectivity.

        Examples
        --------
        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=11)
        >>> point_ids = mesh.get_point_ids([0, 1, 1])
        >>> point_ids
        array([1320])

        >>> cell_ids = mesh.get_cell_ids(point_ids)
        >>> cell_ids
        array([990])

        """
        return np.argwhere(np.isin(self.cells, point_ids).any(axis=1))[:, 0]

    def get_cell_ids_neighbours(self, cell_ids):
        """Return cell ids which share points with given cell ids.

        Parameters
        ----------
        cell_ids : list or ndarray
            Array with cell ids which are used to search for neighbour cells.

        Returns
        -------
        ndarray
            Array with cell ids which are next to the given cells.

        Examples
        --------
        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=11)
        >>> point_ids = mesh.get_point_ids([0, 1, 1])
        >>> point_ids
        array([1320])

        >>> cell_ids = mesh.get_cell_ids(point_ids)
        >>> cell_ids
        array([990])

        Find the cell ids which share at least one point with the given cell id(s).

        >>> cell_ids_neighbours = mesh.get_cell_ids_neighbours(cell_ids)
        >>> cell_ids_neighbours
        array([880, 881, 890, 891, 980, 981, 990, 991])

        """
        return self.get_cell_ids(self.cells[cell_ids])

    def get_point_ids_shared(self, cell_ids_neighbours):
        """Return shared point ids for given cell ids.

        Parameters
        ----------
        cells_neighbours : list or ndarray
            Array with cell ids.

        Returns
        -------
        ndarray
            Array with point ids which are connected to all given cell neighbours.

        Examples
        --------
        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=11)
        >>> point_ids = mesh.get_point_ids([0, 1, 1])
        >>> point_ids
        array([1320])

        >>> cell_ids = mesh.get_cell_ids(point_ids)
        >>> cell_ids
        array([990])

        Find the cell ids which share at least one point with the given cell id(s).

        >>> cell_ids_neighbours = mesh.get_cell_ids_neighbours(cell_ids)
        >>> cell_ids_neighbours
        array([880, 881, 890, 891, 980, 981, 990, 991])

        Find the shared point ids for the list of cell ids.

        >>> point_ids_shared = mesh.get_point_ids_shared(cell_ids_neighbours)
        >>> point_ids_shared
        array([1189])

        """

        neighbours = self.cells[cell_ids_neighbours]
        cell = neighbours[0]
        return cell[[np.isin(neighbours, point).any(axis=1).all() for point in cell]]

    def get_point_ids_corners(self):
        """Return point ids which are located at (xmin, ymin), (xmax, ymin), etc.

        Returns
        -------
        ndarray
            Array with point ids which are located at the corners.

        Examples
        --------
        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=11)
        >>> point_ids_corners = mesh.get_point_ids_corners()
        >>> point_ids_corners
        array([   0, 1210,   10, 1220,  110, 1320,  120, 1330])

        """

        xmin = np.min(self.points, axis=0)
        xmax = np.max(self.points, axis=0)
        points = np.vstack(
            [x.ravel() for x in np.meshgrid(*np.vstack([xmin, xmax]).T)]
        ).T

        return np.concatenate([self.get_point_ids(point) for point in points])

    def modify_corners(self, point_ids=None):
        """Modify the corners of a regular rectangle (quad) or cube (hexahedron)
        inplace. Only the cells array is modified, the points array remains unchanged.

        Parameters
        ----------
        point_ids : ndarray or None, optional
            Array with point ids located at the corners which are modified (default is
            None). If None, all corners are modified.

        Notes
        -----
        Description of the algorithm:

        1. Get corner point ids.

        For each corner point:

        2. Get attached cell and find cell neighours.
        3. Get the shared point of the cell and its neighbours.
        4. Get pair-wise shared points which are located on an edge.
        5. Replace the shared points with the corner point.
        6. Delete the cell attached to the corner point.

        Examples
        --------
        ..  pyvista-plot::
            :context:
            :force_static:

            >>> import numpy as np
            >>> import felupe as fem
            >>>
            >>> mesh = fem.Rectangle(b=(3, 1), n=(16, 6))
            >>> mesh.plot().show()

        ..  pyvista-plot::
            :context:
            :force_static:

            >>> mesh = mesh.modify_corners()  # inplace
            >>> mesh.plot().show()

        """

        if self.cell_type not in ["quad", "hexahedron"]:
            message = [
                "Cell type not supported.",
                "Must be either 'quad' or 'hexahedron'",
                f"but given cell type is '{self.cell_type}'.",
            ]
            raise TypeError(" ".join(message))

        if point_ids is None:
            point_ids = self.get_point_ids_corners()

        for point_id in point_ids:
            cell_id = self.get_cell_ids(point_id)[0]
            cell_id_with_neighbours = self.get_cell_ids_neighbours(cell_id)

            cell_ids_neighbours = cell_id_with_neighbours[
                cell_id_with_neighbours != cell_id
            ]

            point_id_shared = self.get_point_ids_shared(cell_id_with_neighbours)[0]
            point_ids_shared_individual = [
                self.get_point_ids_shared([cell_id, neighbour])
                for neighbour in cell_ids_neighbours
            ]
            if self.cell_type == "hexahedron":
                edges = np.argwhere(
                    np.isclose(self.points, self.points[point_id]).sum(axis=1) >= 2
                )[:, 0]
                point_ids_shared_individual = [
                    p[np.isin(p, edges)] for p in point_ids_shared_individual
                ]
            point_ids_shared_individual = [
                shared[shared != point_id_shared]
                for shared in point_ids_shared_individual
            ]

            for shared, neighbour in zip(
                point_ids_shared_individual, cell_ids_neighbours
            ):
                point_id_replace = np.argwhere(np.isin(self.cells[neighbour], shared))
                if len(point_id_replace) > 0:
                    self.cells[neighbour, point_id_replace[0][0]] = point_id

            self.cells = np.delete(self.cells, cell_id, axis=0)
            self.update(cells=self.cells)

        return self

    def dual(
        self,
        points_per_cell=None,
        disconnect=True,
        calc_points=False,
        offset=0,
        npoints=None,
    ):
        """Create a new dual mesh with given points per cell.

        Parameters
        ----------
        points_per_cell : int or None, optional
            Number of points per cell, must be equal or lower than ``cells.shape[1]`` (
            default is None). If None, all points per cell are considered for the dual
            mesh.
        disconnect : bool, optional
            A flag to disconnect the mesh (each cell has its own points). Default is
            True.
        calc_points : bool, optional
            A flag to calculate the point coordinates for the dual mesh (default is
            False). If False, the points array is filled with zeros.
        offset : int, optional
            An offset to be added to the cells array (default is 0).
        npoints : int or None, optional
            Number of points for the dual mesh. If the given number of points is greater
            than ``npoints * points_per_cell``, then the missing points are added to the
            points array (filled with zeros). Default is None.

        Returns
        -------
        Mesh
            The dual mesh.

        Notes
        -----
        ..  note::
            The points array of the dual mesh always has a shape of
            ``(npoints * points_per_cell, dim)``.

        Examples
        --------
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle(n=5).add_midpoints_edges()
        >>> region = fem.RegionQuadraticQuad(mesh=mesh)
        >>>
        >>> mesh_dual = mesh.dual(points_per_cell=1, disconnect=False)
        >>> region_dual = fem.RegionConstantQuad(
        ...     mesh_dual, quadrature=region.quadrature, grad=False
        ... )
        >>>
        >>> displacement = fem.FieldPlaneStrain(region, dim=2)
        >>> pressure = fem.Field(region_dual)
        >>> field = fem.FieldContainer([displacement, pressure])

        See Also
        --------
        felupe.mesh.dual : Create a new dual mesh with given points per cell.

        """
        return as_mesh(
            dual(
                self,
                points_per_cell=points_per_cell,
                disconnect=disconnect,
                calc_points=calc_points,
                offset=offset,
                npoints=npoints,
            )
        )

    def expand(self, n=11, z=1, axis=-1, expand_dim=True):
        """Expand a 0d-Point to a 1d-Line, a 1d-Line to a 2d-Quad or a 2d-Quad to a
        3d-Hexahedron Mesh.

        Parameters
        ----------
        n : int, optional
            Number of n-point repetitions or (n-1)-cell repetitions, default is 11. Must
            be greater than 0.
        z : float or ndarray, optional
            Total expand dimension as float (edge length in expand direction is z / n),
            default is 1. Optionally, if an array is passed these entries are taken as
            expansion and ``n`` is ignored.
        axis : int, optional
            Axis of expansion (default is -1).
        mask : ndarray or None, optional
            A boolean mask to select points which are rotated (default is None).
        expand_dim : bool, optional
            Expand the dimension of the point coordinates (default is True).

        Returns
        -------
        Mesh
            The expanded mesh.

        Examples
        --------
        Expand a rectangle to a cube.

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> rect = fem.Rectangle(n=4)
           >>> cube = rect.expand(n=7, z=2)
           >>>
           >>> cube.plot().show()

        >>> cube
        <felupe Mesh object>
          Number of points: 112
          Number of cells:
            hexahedron: 54

        See Also
        --------
        felupe.mesh.expand : Expand a 0d-Point to a 1d-Line, a 1d-Line to a 2d-Quad or a
            2d-Quad to a 3d-Hexahedron Mesh.
        """
        return as_mesh(expand(self, n=n, z=z, axis=axis, expand_dim=expand_dim))

    def rotate(self, angle_deg, axis, center=None, mask=None):
        """Rotate a Mesh.

        Parameters
        ----------
        angle_deg : int
            Rotation angle in degree.
        axis : int
            Rotation axis.
        center : list or ndarray or None, optional
            Center point coordinates (default is None).
        mask : ndarray or None, optional
            A boolean mask to select points which are rotated (default is None).

        Returns
        -------
        Mesh
            The rotated mesh.

        Examples
        --------
        Rotate a rectangle in the xy-plane by 35 degree.

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> rect = fem.Rectangle(b=(3, 1), n=(10, 4))
           >>> mesh = rect.rotate(angle_deg=35, axis=2, center=[1.5, 0.5])
           >>> mesh.plot().show()

        >>> mesh
        <felupe Mesh object>
          Number of points: 40
          Number of cells:
            quad: 27

        See Also
        --------
        felupe.mesh.rotate : Rotate a Mesh.
        """
        return as_mesh(
            rotate(self, angle_deg=angle_deg, axis=axis, center=center, mask=mask)
        )

    def revolve(self, n=11, phi=180, axis=0, expand_dim=True):
        """Revolve a 2d-Quad to a 3d-Hexahedron Mesh.

        Parameters
        ----------
        n : int, optional
            Number of n-point revolutions (or (n-1) cell revolutions),
            default is 11.
        phi : float or ndarray, optional
            Revolution angle in degree (default is 180).
        axis : int, optional
            Revolution axis (default is 0).
        expand_dim : bool, optional
            Expand the dimension of the point coordinates (default is True).

        Returns
        -------
        Mesh
            The revolved mesh.

        Examples
        --------
        Revolve a cylinder from a rectangle.

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> rect = fem.Rectangle(a=(0, 4), b=(3, 5), n=(10, 4))
           >>> mesh = rect.revolve(n=11, phi=180, axis=0)
           >>> mesh.plot().show()

        >>> mesh
        <felupe Mesh object>
          Number of points: 440
          Number of cells:
            hexahedron: 270

        See Also
        --------
        felupe.mesh.revolve : Revolve a 2d-Quad to a 3d-Hexahedron Mesh.
        """
        return as_mesh(revolve(self, n=n, phi=phi, axis=axis, expand_dim=expand_dim))

    def merge_duplicate_points(self, decimals=None):
        """Merge duplicate points and update cells of a Mesh.

        Parameters
        ----------
        decimals : int or None, optional
            Number of decimals for point coordinate comparison (default is None).

        Returns
        -------
        Mesh
            The mesh with merged duplicate points and updated cells.

        Notes
        -----
        ..  warning::
            This function re-sorts points.

        ..  note::
            This function does not merge duplicate cells.

        Examples
        --------
        Two quad meshes to be merged overlap some points. Merge these duplicated
        points and update the cells.

        ..  pyvista-plot::
            :context:

            >>> import felupe as fem
            >>>
            >>> rect1 = fem.Rectangle(n=11)
            >>> rect2 = fem.Rectangle(a=(0.9, 0), b=(1.9, 1), n=11)
            >>> rect2
            <felupe Mesh object>
              Number of points: 121
              Number of cells:
                quad: 100

        Each mesh contains 121 points and 100 cells. These two meshes are now stored in a
        :class:`~felupe.MeshContainer`.

        ..  pyvista-plot::
            :context:

            >>> container = fem.MeshContainer([rect1, rect2])
            >>> container
            <felupe mesh container object>
              Number of points: 242
              Number of cells:
                quad: 100
                quad: 100

        The meshes of the mesh container are :func:`stacked <felupe.mesh.stack>`.

        ..  pyvista-plot::
            :context:

            >>> stack = fem.mesh.stack(container.meshes)
            >>> stack
            <felupe Mesh object>
              Number of points: 242
              Number of cells:
                quad: 200

        After merging the duplicated points and cells, the number of points is reduced
        but the number of cells is unchanged.

        ..  pyvista-plot::
            :context:
            :force_static:

            >>> mesh = stack.merge_duplicate_points()
            >>> mesh
            <felupe Mesh object>
              Number of points: 220
              Number of cells:
                quad: 200

            >>> mesh.plot(opacity=0.6).show()

        ..  note::
            The :class:`~felupe.MeshContainer` may be directly created with
            ``merge=True``. This enforces :func:`~felupe.mesh.merge_duplicate_points`
            for the shared points array of the container.

        See Also
        --------
        felupe.mesh.merge_duplicate_points : Merge duplicated points and update cells of
            a Mesh.
        felupe.MeshContainer : A container which operates on a list of meshes with
            identical dimensions.
        """
        return as_mesh(merge_duplicate_points(self, decimals=decimals))

    @wraps(merge_duplicate_cells)
    def merge_duplicate_cells(self):
        """Merge duplicate cells of a Mesh.

        Returns
        -------
        Mesh
            The mesh with merged duplicate cells.

        Notes
        -----
        ..  warning::
            This function re-sorts cells.

        ..  note::
            This function does not merge duplicate points.

        Examples
        --------
        Two quad meshes to be merged overlap some cells. Merge these duplicated
        points and update the cells.

        ..  pyvista-plot::
            :context:

            >>> import felupe as fem
            >>>
            >>> rect1 = fem.Rectangle(n=11)
            >>> rect2 = fem.Rectangle(a=(0.9, 0), b=(1.9, 1), n=11)
            >>> rect2
            <felupe Mesh object>
              Number of points: 121
              Number of cells:
                quad: 100

        Each mesh contains 121 points and 100 cells. These two meshes are now stored in
        a :class:`~felupe.MeshContainer`.

        ..  pyvista-plot::
            :context:

            >>> container = fem.MeshContainer([rect1, rect2])
            >>> container
            <felupe mesh container object>
              Number of points: 242
              Number of cells:
                quad: 100
                quad: 100

        The meshes of the mesh container are :func:`stacked <felupe.mesh.stack>`.

        ..  pyvista-plot::
            :context:

            >>> stack = fem.mesh.stack(container.meshes)
            >>> stack
            <felupe Mesh object>
              Number of points: 242
              Number of cells:
                quad: 200

        After merging the duplicated points and cells, the number of points is reduced
        but the number of cells is unchanged.

        ..  pyvista-plot::
            :context:
            :force_static:

            >>> mesh = stack.merge_duplicate_points()
            >>> mesh
            <felupe Mesh object>
              Number of points: 220
              Number of cells:
                quad: 200

            mesh.plot(opacity=0.6).show()

        ..  note::
            The :class:`~felupe.MeshContainer` may be directly created with
            ``merge=True``. This enforces :func:`~felupe.mesh.merge_duplicate_points`
            for the shared points array of the container.

        The duplicate cells are merged in a second step.

        ..  pyvista-plot::
            :context:

            >>> merged = mesh.merge_duplicate_cells()
            >>> merged
            <felupe Mesh object>
              Number of points: 220
              Number of cells:
                quad: 190

        See Also
        --------
        felupe.mesh.merge_duplicate_points : Merge duplicate points of a Mesh.
        felupe.mesh.merge_duplicate_cells : Merge duplicate cells of a Mesh.
        felupe.MeshContainer : A container which operates on a list of meshes with
            identical dimensions.
        """
        return as_mesh(merge_duplicate_cells(self))

    def fill_between(self, other_mesh, n=11):
        """Fill a 2d-Quad Mesh between two 1d-Line Meshes, embedded in 2d-space, or a
        3d-Hexahedron Mesh between two 2d-Quad Meshes, embedded in 3d-space, by expansion.
        Both meshes must have equal number of points and cells. The cells-array is taken
        from the first mesh.

        Parameters
        ----------
        other_mesh : felupe.Mesh
            The other line- or quad-mesh.
        n : int or ndarray
            Number of n-point repetitions or (n-1)-cell repetitions,
            (default is 11). If an array is given, then its values are used for the
            relative positions in a reference configuration (-1, 1) between the two meshes.

        Returns
        -------
        felupe.Mesh
            The expanded mesh.

        Examples
        --------
        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> inner = fem.mesh.revolve(fem.Point(1)).expand(z=0.4).translate(0.2, axis=2)
           >>> outer = fem.mesh.revolve(fem.Point(2), phi=160).rotate(
           ...     axis=2, angle_deg=20
           ... ).expand(z=1.2)
           >>> mesh = inner.fill_between(outer, n=6)
           >>>
           >>> mesh.plot().show()

        See Also
        --------
        felupe.mesh.fill_between : Fill a 2d-Quad Mesh between two 1d-Line Meshes, embedded
            in 2d-space, or a 3d-Hexahedron Mesh between two 2d-Quad Meshes, embedded in
            3d-space, by expansion.

        """
        return as_mesh(fill_between(self, other_mesh=other_mesh, n=n))

    def flip(self, mask=None):
        r"""Ensure positive cell volumes for `tria`, `tetra`, `quad` and `hexahedron`
        cell types.

        Parameters
        ----------
        mask: list, ndarray or None, optional
            Boolean mask for selected cells to flip (default is None). If None, all
            cells are selected to be flipped.

        Returns
        -------
        Mesh
            The mesh with a rearranged cells array to ensure positive cell volumes.

        Examples
        --------
        A quad mesh with negative cell volumes occurs if one coordinate axis is
        multiplied by -1. The error pops up if a region is created with this mesh.

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle(n=3)
        >>> mesh.update(points=mesh.points * np.array([[-1, 1]]))
        >>> region = fem.RegionQuad(mesh)

        The sum of the differential volumes :math:`V = \sum_c \sum_q dV_{qc}` is
        evaluated to -1.0.

        >>> region.dV.sum()
        -1.0

        Let's try to fix the mesh.

        >>> mesh.cells
        array([[0, 1, 4, 3],
               [1, 2, 5, 4],
               [3, 4, 7, 6],
               [4, 5, 8, 7]])

        The cells array is rearranged to ensure positive cell volumes.

        >>> mesh_fixed = mesh.flip()
        >>> mesh_fixed.cells
        array([[3, 4, 1, 0],
               [4, 5, 2, 1],
               [6, 7, 4, 3],
               [7, 8, 5, 4]])

        A region now correctly evaluates the total volume of the mesh to 1.0.

        >>> region_fixed = fem.RegionQuad(mesh_fixed)
        >>> region_fixed.dV.sum()
        1.0

        See Also
        --------
        felupe.mesh.flip : Ensure positive cell volumes for `tria`, `tetra`, `quad` and
            `hexahedron` cell types.

        """
        return as_mesh(flip(self, mask=mask))

    def mirror(self, normal=[1, 0, 0], centerpoint=[0, 0, 0], axis=None):
        """Mirror points by plane normal and ensure positive cell volumes for
        `tria`, `tetra`, `quad` and `hexahedron` cell types.

        Parameters
        ----------
        normal: list or ndarray, optional
            Mirror-plane normal vector (default is [1, 0, 0]).
        centerpoint: list or ndarray, optional
            Center-point coordinates on the mirror plane (default is [0, 0, 0]).
        axis: int or None, optional
            Mirror axis (default is None).

        Returns
        -------
        Mesh
            The mirrored mesh.

        Examples
        --------
        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> mesh = fem.Circle(sections=[0, 90, 180], n=5)
           >>> mesh.plot().show()

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> mesh = fem.Circle(sections=[0, 90, 180], n=5)
           >>> mesh.mirror(normal=[0, 1, 0]).plot().show()

        See Also
        --------
        felupe.Mesh.mirror : Mirror points by plane normal and ensure positive cell
            volumes for `tria`, `tetra`, `quad` and `hexahedron` cell types.
        """
        return as_mesh(mirror(self, normal=normal, centerpoint=centerpoint, axis=axis))

    def translate(self, move, axis):
        """Translate (move) a Mesh along a given axis.

        Parameters
        ----------
        move : float
            Translation along given axis.
        axis : int
            Translation axis.

        Returns
        -------
        Mesh
            The translated mesh.

        Examples
        --------
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Circle(n=6)
        >>> mesh.points.min(axis=0), mesh.points.max(axis=0)
        (array([-1., -1.]), array([1., 1.]))

        >>> translated = mesh.translate(0.3, axis=1)
        >>> translated.points.min(axis=0), translated.points.max(axis=0)
        (array([-1. , -0.7]), array([1. , 1.3]))

        See Also
        --------
        felupe.mesh.translate : Translate (move) a Mesh along a given axis.
        """

        return as_mesh(translate(self, move=move, axis=axis))

    def triangulate(self, mode=3):
        """Triangulate a quad or a hex mesh.

        Parameters
        ----------
        mode: int, optional
            Choose a mode how to convert hexahedrons to tets [1]_ (default is 3).

        Returns
        -------
        Mesh
            The triangulated mesh.

        Examples
        --------
        Use ``mode=0`` to convert a mesh of hexahedrons into tetrahedrons [1]_.

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> mesh = fem.Cube(n=6)
           >>> mesh.triangulate(mode=0).plot().show()

        Use ``mode=3`` to convert a mesh of hexahedrons into tetrahedrons [1]_.

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> mesh = fem.Cube(n=6)
           >>> mesh.triangulate(mode=3).plot().show()

        References
        ----------
        .. [1] Dompierre, J., Labbé, P., Vallet, M. G., & Camarero, R. (1999).
           How to Subdivide Pyramids, Prisms, and Hexahedra into Tetrahedra.
           IMR, 99, 195.

        See Also
        --------
        felupe.mesh.triangulate : Triangulate a quad or a hex mesh.
        """
        return as_mesh(triangulate(self, mode=mode))

    def add_runouts(
        self,
        values=[0.1, 0.1],
        centerpoint=[0, 0, 0],
        axis=0,
        exponent=5,
        mask=slice(None),
        normalize=False,
    ):
        """Add simple rubber-runouts for realistic rubber-metal structures.

        Parameters
        ----------
        values : list or ndarray, optional
            Relative amount of runouts (per coordinate) perpendicular to the axis
            (default is 10% per coordinate, i.e. [0.1, 0.1]).
        centerpoint : list or ndarray, optional
            Center-point coordinates (default is [0, 0, 0]).
        axis : int or None, optional
            Axis (default is 0).
        exponent : int, optional
            Positive exponent to control the shape of the runout. The higher
            the exponent, the steeper the transition (default is 5).
        mask : list or None, optional
            List of points to be considered (default is None).
        normalize : bool, optional
            Normalize the runouts to create indents, i.e. maintain the original shape at the
            ends (default is False).

        Returns
        -------
        Mesh
            The mesh with the modified point coordinates.

        Examples
        --------
        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> rect = fem.Rectangle(a=(-3, -1), b=(3, 1), n=(31, 11))
           >>> mesh = rect.add_runouts(axis=1, values=[0.2], normalize=True)
           >>>
           >>> mesh.plot().show()

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> cube = fem.Cube(a=(-3, -2, -1), b=(3, 2, 1), n=(31, 21, 11))
           >>> mesh = cube.add_runouts(axis=2, values=[0.1, 0.3], normalize=True)
           >>>
           >>> mesh.plot().show()

        See Also
        --------
        felupe.mesh.runouts : Add simple rubber-runouts for realistic rubber-metal
            structures.

        """
        return as_mesh(
            runouts(
                self,
                values=values,
                centerpoint=centerpoint,
                axis=axis,
                exponent=exponent,
                mask=mask,
                normalize=normalize,
            )
        )

    def convert(
        self,
        order=0,
        calc_points=False,
        calc_midfaces=False,
        calc_midvolumes=False,
    ):
        """Convert a mesh to a given order. Only conversions to ``order=0`` and
        ``order=2`` are supported. This function supports meshes with cell types
        ``"triangle"``, ``"tetra"``, ``"quad"`` and ``"hexahedron"``.

        Parameters
        ----------
        order : int, optional
            The order of the converted mesh (default is 0). If 0, the points-array will
            be of shape ``(ncells, dim)``. If 0 and ``calc_points`` is True, the mean of
            all points per cell is evaluated. If 0 and ``calc_points`` is False, the
            points array is filled with zeros. If 2, at least midpoints on cell edges
            are added to the mesh. If 2 and ``calc_midfaces`` is True, midpoints on cell
            faces are also added. If 2 and ``calc_midvolumes`` is True, midpoints on
            cell volumes are also added. Raises an error if not 0 or 2.
        calc_points : bool, optional
            Flag to return the mean of all points per cell if ``order=0`` (default is
            False). If False, the points-array is filled with zeros.
        calc_midfaces : bool, optional
            Flag to add midpoints on cell faces if ``order=2`` (default is False).
        calc_midvolumes : bool, optional
            Flag to add midpoints on cell volumes if ``order=2`` (default is False).

        Returns
        -------
        Mesh
            The converted mesh.

        Examples
        --------
        Convert a mesh of hexahedrons to quadratic hexahedrons by inserting midpoints on
        the cell edges.

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> mesh = fem.Rectangle(n=6)
           >>> mesh2 = mesh.convert(order=2)
           >>>
           >>> mesh2.plot(plotter=mesh.plot(), style="points", color="black").show()

        >>> mesh2
        <felupe Mesh object>
          Number of points: 96
          Number of cells:
            quad8: 25

        See Also
        --------
        felupe.mesh.add_midpoints_edges : Add midpoints on edges for given points and cells
            and update cell_type accordingly.
        felupe.mesh.add_midpoints_faces : Add midpoints on faces for given points and cells
            and update cell_type accordingly.
        felupe.mesh.add_midpoints_volumes : Add midpoints on volumes for given points and
            cells and update cell_type accordingly.
        felupe.Mesh.add_midpoints_edges : Add midpoints on edges for given points and cells
            and update cell_type accordingly.
        felupe.Mesh.add_midpoints_faces : Add midpoints on faces for given points and cells
            and update cell_type accordingly.
        felupe.Mesh.add_midpoints_volumes : Add midpoints on volumes for given points and
            cells and update cell_type accordingly.

        """
        return as_mesh(
            convert(
                self,
                order=order,
                calc_points=calc_points,
                calc_midfaces=calc_midfaces,
                calc_midvolumes=calc_midvolumes,
            )
        )

    @wraps(collect_edges)
    def collect_edges(self):
        return collect_edges(self)

    @wraps(collect_faces)
    def collect_faces(self):
        return collect_faces(self)

    @wraps(collect_volumes)
    def collect_volumes(self):
        return collect_volumes(self)

    def add_midpoints_edges(self, cell_type=None):
        """Add midpoints on edges for given points and cells and update cell_type
        accordingly.

        Parameters
        ----------
        cell_type: str or None, optional
            A string in VTK-convention that specifies the new cell type (default is None).
            If None, the cell type is chosen automatically.

        Returns
        -------
        Mesh
            A new mesh with inserted midpoints on cell edges.

        Examples
        --------
        Convert a mesh of hexahedrons to quadratic hexahedrons by inserting midpoints on
        the cell edges.

        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> mesh = fem.Rectangle(n=6)
           >>> mesh_with_midpoints_edges = mesh.add_midpoints_edges()
           >>>
           >>> mesh_with_midpoints_edges.plot(
           ...     plotter=mesh.plot(), style="points", color="black"
           ... ).show()

        >>> mesh_with_midpoints_edges
        <felupe Mesh object>
          Number of points: 96
          Number of cells:
            quad8: 25

        See Also
        --------
        felupe.mesh.add_midpoints_edges : Add midpoints on edges for given points and
            cells and update cell_type accordingly.

        """
        return add_midpoints_edges(self, cell_type_new=cell_type)

    def add_midpoints_faces(self, cell_type=None):
        """Add midpoints on faces for given points and cells and update cell_type
        accordingly.

        Parameters
        ----------
        cell_type: str or None, optional
            A string in VTK-convention that specifies the new cell type (default is
            None). If None, the cell type is chosen automatically.

        Returns
        -------
        Mesh
            A new mesh with inserted midpoints on cell faces.

        Examples
        --------
        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>>
           >>> mesh = fem.Rectangle(n=6)
           >>> mesh_with_midpoints_faces = mesh.add_midpoints_faces(cell_type="quad")
           >>>
           >>> mesh_with_midpoints_faces.plot(
           ...     plotter=mesh.plot(), style="points", color="black"
           ... ).show()

        >>> mesh_with_midpoints_faces
        <felupe Mesh object>
          Number of points: 61
          Number of cells:
            quad: 25

        See Also
        --------
        felupe.mesh.add_midpoints_faces : Add midpoints on faces for given points and
            cells and update cell_type accordingly.
        """
        return add_midpoints_faces(self, cell_type_new=cell_type)

    def add_midpoints_volumes(self, cell_type=None):
        """Add midpoints on volumes for given points and cells and update cell_type
        accordingly.

        Parameters
        ----------
        cell_type: str or None, optional
            A string in VTK-convention that specifies the new cell type (default is None).
            If None, the cell type is chosen automatically.

        Returns
        -------
        Mesh
            A new mesh with inserted midpoints on cell volumes.

        Examples
        --------
        .. pyvista-plot::
           :force_static:

           >>> import felupe as fem
           >>> import pyvista as pv
           >>>
           >>> mesh = fem.Cube(n=6)
           >>> mesh_with_midpoints_volumes = fem.mesh.add_midpoints_volumes(
           ...     mesh, cell_type_new="hexahedron9"
           ... )
           >>> plotter = pv.Plotter()
           >>> actor = plotter.add_points(mesh_with_midpoints_volumes.points, color="black")
           >>> mesh.plot(opacity=0.5, plotter=plotter).show()

        >>> mesh_with_midpoints_volumes
        <felupe Mesh object>
          Number of points: 341
          Number of cells:
            hexahedron9: 125

        See Also
        --------
        felupe.mesh.add_midpoints_volumes : Add midpoints on volumes for given points
            and cells and update cell_type accordingly.
        """
        return add_midpoints_volumes(self, cell_type_new=cell_type)

    def subdivide(self):
        """Subdivide cells by adding midpoints on edges, faces and volumes.

        Returns
        -------
        Mesh
            A new subdivided mesh.

        Examples
        --------
        Take a rectangle mesh and subdivide it two times.

        .. pyvista-plot::
           :context:
           :force_static:

           >>> import felupe as fem
           >>>
           >>> rect = fem.Rectangle(n=4).modify_corners()
           >>> rect.plot().show()

        .. pyvista-plot::
           :context:
           :force_static:

           >>> mesh = rect.subdivide().subdivide()
           >>> mesh.plot().show()

        See Also
        --------
        felupe.Mesh.subdivide : Subdivide cells by adding midpoints on edges, faces and
            volumes.
        """
        return subdivide(self)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ...math import cdya_ik, cdya_il, det, dot, dya, identity, inv, trace, transpose
from .._base import ConstitutiveMaterial


class NeoHookeCompressible(ConstitutiveMaterial):
    r"""Compressible isotropic hyperelastic Neo-Hookean material formulation. The strain
    energy density function of the Neo-Hookean material formulation is a linear function
    of the trace of the right Cauchy-Green deformation tensor.

    Parameters
    ----------
    mu : float or None, optional
        Shear modulus (second Lamé constant). Default is None.
    lmbda : float or None, optional
        First Lamé constant (default is None)

    Notes
    -----

    .. math::

        \psi &= \psi(\boldsymbol{C})

        \psi(\boldsymbol{C}) &= \frac{\mu}{2} \text{tr}(\boldsymbol{C})
            - \mu \ln(J) + \frac{\lambda}{2} \ln(J)^2

    with

    .. math::

       J = \text{det}(\boldsymbol{F})

    The first Piola-Kirchhoff stress tensor is evaluated as the gradient
    of the strain energy density function.

    .. math::

       \boldsymbol{P} &= \frac{\partial \psi}{\partial \boldsymbol{F}}

       \boldsymbol{P} &= \mu \left( \boldsymbol{F} - \boldsymbol{F}^{-T} \right)
           + \lambda \ln(J) \boldsymbol{F}^{-T}

    The hessian of the strain energy density function enables the corresponding
    elasticity tensor.

    .. math::

       \mathbb{A} &= \frac{\partial^2 \psi}{\partial \boldsymbol{F}\ \partial
       \boldsymbol{F}}

       \mathbb{A} &= \mu \boldsymbol{I} \overset{ik}{\otimes} \boldsymbol{I}
           + \left(\mu - \lambda \ln(J) \right)
               \boldsymbol{F}^{-T} \overset{il}{\otimes} \boldsymbol{F}^{-T}
           + \lambda \boldsymbol{F}^{-T} {\otimes} \boldsymbol{F}^{-T}

    Examples
    --------
    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.NeoHookeCompressible(mu=1.0, lmbda=2.0)
        >>> ax = umat.plot()

    """

    def __init__(self, mu=None, lmbda=None, parallel=False):
        self.parallel = parallel

        self.mu = mu
        self.lmbda = lmbda

        self.kwargs = {"mu": self.mu}
        if self.lmbda is not None:
            self.kwargs["lmbda"] = self.lmbda

        # aliases for function, gradient and hessian
        self.energy = self.function
        self.stress = self.gradient
        self.elasticity = self.hessian

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [np.eye(3), np.zeros(0)]

    def function(self, x):
        """Strain energy density function per unit undeformed volume of the Neo-Hookean
        material formulation.

        Parameters
        ----------
        x : list of ndarray
            List with the Deformation gradient ``F`` (3x3) as first item
        """

        F = x[0]

        mu = self.mu
        lmbda = self.lmbda

        lnJ = np.log(det(F))
        C = dot(transpose(F), F, parallel=self.parallel)

        W = mu * (trace(C) / 2 - lnJ)

        if lmbda is not None:
            W += lmbda * lnJ**2 / 2

        return [W]

    def gradient(self, x, out=None):
        """Gradient of the strain energy density function per unit undeformed volume of
        the Neo-Hookean material formulation.

        Parameters
        ----------
        x : list of ndarray
            List with the Deformation gradient ``F`` (3x3) as first item
        out : ndarray or None, optional
            A location into which the result is stored (default is None).
        """

        F, statevars = x[0], x[-1]

        mu = self.mu
        lmbda = self.lmbda

        J = det(F)
        iFT = transpose(inv(F, J))
        lnJ = np.log(J, out=J)

        P = np.multiply(mu, F, out=out)

        if lmbda is None:
            Pb = np.multiply(iFT, -mu, out=iFT)
        else:
            lmbda_lnJ = np.multiply(lmbda, lnJ, out=lnJ)
            Pb = np.multiply(iFT, -mu + lmbda_lnJ, out=iFT)

        np.add(P, Pb, out=P)

        return [P, statevars]

    def hessian(self, x, out=None):
        """Hessian of the strain energy density function per unit undeformed volume of
        the Neo-Hookean material formulation.

        Parameters
        ----------
        x : list of ndarray
            List with the Deformation gradient ``F`` (3x3) as first item
        out : ndarray or None, optional
            A location into which the result is stored (default is None).
        """

        F = x[0]

        mu = self.mu
        lmbda = self.lmbda

        J = det(F)
        iFT = transpose(inv(F, J))
        lnJ = np.log(J, out=J)
        eye = identity(F)

        iFTiFT = cdya_il(iFT, iFT, out=out)
        A4a = cdya_ik(eye, eye)
        np.multiply(mu, A4a, out=A4a)

        if lmbda is not None:
            lmbda_lnJ = np.multiply(lmbda, lnJ, out=lnJ)
            A4b = np.multiply(mu - lmbda_lnJ, iFTiFT, out=iFTiFT)
            iFT_iFT = dya(iFT, iFT)
            A4c = np.multiply(lmbda, iFT_iFT, out=iFT_iFT)
            np.add(A4a, A4b, out=A4b)
            A4 = np.add(A4b, A4c, out=A4c)
        else:
            A4b = np.multiply(mu, iFTiFT, out=iFTiFT)
            A4 = np.add(A4a, A4b, out=A4b)

        return [A4]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import trace
from tensortrax.math.linalg import det, inv
from tensortrax.math.special import from_triu_1d, triu_1d


def finite_strain_viscoelastic(C, Cin, mu, eta, dtime):
    r"""Multiplicative
    `finite strain viscoelastic <https://doi.org/10.1016/j.cma.2013.07.004>`_ [1]_
    material formulation.

    Notes
    -----
    The material formulation is built upon the multiplicative decomposition of the
    distortional part of the deformation gradient tensor into an elastic and an
    inelastic part, see Eq. :eq:`multiplicative-split`.

    ..  math::
        :label: multiplicative-split

        \hat{\boldsymbol{F}} &= \boldsymbol{F}_e \boldsymbol{F}_i
            \ (= \hat{\boldsymbol{F}_e} \boldsymbol{F}_i)

        \boldsymbol{C}_e &= \boldsymbol{F}_e^T \boldsymbol{F}_e

        \boldsymbol{C}_i &= \boldsymbol{F}_i^T \boldsymbol{F}_i

        \text{tr}\left( \boldsymbol{C}_e \right) &= \text{tr}\left(
            \hat{\boldsymbol{C}} \boldsymbol{C}_i^{-1}
        \right)

    The components of the inelastic right Cauchy-Green deformation tensor are used as
    state variables with the evolution equation and its explicit update formula as given
    in Eq. :eq:`evolution` [1]_. The elastic part of the multiplicative decomposition of
    the deformation gradient tensor is also enforced to be an unimodular tensor which
    leads to the constraint :math:`\det(\boldsymbol{F_i})=1`. Hence, the inelastic right
    Cauchy-Green deformation tensor must be an unimodular tensor
    :math:`\det(\boldsymbol{C_i})=1`.

    ..  math::
        :label: evolution

        \dot{\boldsymbol{C}}_i &= \frac{\mu}{\eta}\ \hat{\boldsymbol{C}}

        \boldsymbol{C}_i &= \hat{\overline{\boldsymbol{C}_{i,n}
            + \frac{\Delta t \mu}{\eta} \hat{\boldsymbol{C}}}}

    The distortional part of the strain energy density per unit undeformed volume is
    assumed to be of a Neo-Hookean form, see Eq. :eq:`nh-w`.

    ..  math::
        :label: nh-w

        \hat{\psi} = \frac{\mu}{2} \left( \text{tr}\left(
            \hat{\boldsymbol{C}} \boldsymbol{C}_i^{-1}
        \right) - 3 \right)

    Examples
    --------
    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.Hyperelastic(
        ...     fem.finite_strain_viscoelastic, mu=1.0, eta=1.0, dtime=1.0, nstatevars=6
        ... )
        >>> ax = umat.plot(
        ...    incompressible=True,
        ...    ux=fem.math.linsteps([1, 1.5, 1, 2, 1], num=[5, 5, 10, 10]),
        ...    ps=None,
        ...    bx=None,
        ... )

    References
    ----------
    ..  [1] A. V. Shutov, R. Landgraf, and J. Ihlemann, "An explicit solution for
        implicit time stepping in multiplicative finite strain viscoelasticity",
        Computer Methods in Applied Mechanics and Engineering, vol. 265. Elsevier BV,
        pp. 213–225, Oct. 2013. doi:
        `10.1016/j.cma.2013.07.004 <https://doi.org/10.1016/j.cma.2013.07.004>`_.

    """
    J3 = det(C) ** (-1 / 3)

    # update of state variables by evolution equation
    Ci = from_triu_1d(Cin[:6], like=C) + (mu / eta * dtime) * (J3 * C)
    Ci = det(Ci) ** (-1 / 3) * Ci

    # first invariant of elastic part of right Cauchy-Green deformation tensor
    I1 = J3 * trace(C @ inv(Ci))

    # strain energy function and state variable
    return mu / 2 * (I1 - 3), triu_1d(Ci)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import trace
from tensortrax.math.linalg import det


def third_order_deformation(C, C10, C01, C11, C20, C30):
    r"""Strain energy function of the isotropic hyperelastic
    `Third-Order-Deformation <https://doi.org/10.1002/app.1975.070190723>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    C10 : float
        Material parameter associated to the linear term of the first invariant.
    C01 : float
        Material parameter associated to the linear term of the second invariant.
    C11 : float
        Material parameter associated to the mixed term of the first and second
        invariant.
    C20 : float
        Material parameter associated to the quadratic term of the first invariant.
    C30 : float
        Material parameter associated to the cubic term of the first invariant.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-tod`

    ..  math::
        :label: psi-tod

        \psi &= C_{10} \left(\hat{I}_1 - 3 \right) + C_{01} \left(\hat{I}_2 - 3 \right)
             + C_{11} \left(\hat{I}_1 - 3 \right) \left(\hat{I}_2 - 3 \right)

            &+ C_{20} \left(\hat{I}_1 - 3 \right)^2
             + C_{30} \left(\hat{I}_1 - 3 \right)^3

    with the first and second main invariant of the distortional part of the right
    Cauchy-Green deformation tensor, see Eq. :eq:`invariants-tod`.

    ..  math::
        :label: invariants-tod

        \hat{I}_1 &= J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

        \hat{I}_2 &= J^{-4/3} \frac{1}{2} \left(
            \text{tr}\left(\boldsymbol{C}\right)^2 -
            \text{tr}\left(\boldsymbol{C}^2\right)
        \right)

    The doubled sum of the material parameters :math:`C_{10}` and :math:`C_{01}` is
    equal to the initial shear modulus :math:`\mu` as denoted in Eq.
    :eq:`shear-modulus-tod`.

    ..  math::
        :label: shear-modulus-tod

        \mu = 2 \left( C_{10} + C_{01} \right)

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.third_order_deformation,
        ...     C10=0.5,
        ...     C01=0.1,
        ...     C11=0.01,
        ...     C20=-0.1,
        ...     C30=0.02,
        ... )
        >>> ax = umat.plot(incompressible=True)

    """
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = (I1**2 - J3**2 * trace(C @ C)) / 2
    return (
        C10 * (I1 - 3)
        + C01 * (I2 - 3)
        + C11 * (I1 - 3) * (I2 - 3)
        + C20 * (I1 - 3) ** 2
        + C30 * (I1 - 3) ** 3
    )



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from .._base import ConstitutiveMaterial


class LinearElastic1D(ConstitutiveMaterial):
    r"""Isotropic one-dimensional linear-elastic material formulation.

    Parameters
    ----------
    E : float
        Young's modulus.

    Notes
    -----
    The stress-stretch relation of the linear-elastic material formulation is given in
    Eq. :eq:`linear-elastic-1d`

    ..  math::
        :label: linear-elastic-1d

        \sigma = E\ \left(\lambda -  1 \right)

    with the stretch from Eq. :eq:`linear-elastic-strain-1d`.

    ..  math::
        :label: linear-elastic-strain-1d

        \lambda = \frac{l}{L}


    Examples
    --------
    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.LinearElastic1D(E=1)
        >>> ax = umat.plot()

    """

    def __init__(self, E):
        self.E = np.array(E)

        self.kwargs = {"E": self.E}

        # aliases for gradient and hessian
        self.stress = self.gradient
        self.elasticity = self.hessian

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [np.ones(1), np.zeros(0)]

    def gradient(self, x, out=None):
        r"""Evaluate the stress (as a function of the stretch).

        Parameters
        ----------
        x : list of ndarray
            List with the stretch :math:`\lambda` as first item.
        out : ndarray or None, optional
            A location into which the result is stored (default is None).
            Not implemented.

        Returns
        -------
        ndarray
            Stress

        """

        λ, statevars = x[0], x[-1]

        return [self.E * (λ - 1), statevars]

    def hessian(self, x=None, shape=(1,), dtype=None, out=None):
        r"""Evaluate the elasticity. The stretch is only used for the shape of the
        trailing axes.

        Parameters
        ----------
        x : list of ndarray, optional
            List with stretch :math:`\lambda` as first item (default is None).
        shape : tuple of int, optional
            Tuple with shape of the trailing axes (default is (1,)).
        out : ndarray or None, optional
            A location into which the result is stored (default is None).
            Not implemented.

        Returns
        -------
        ndarray
            Elasticity

        """

        return [self.E]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


def solve_nd(A, b, solve=np.linalg.solve, n=1, **kwargs):
    r"""Solve a linear equation system for n-dimensional unknowns with optional
    elementwise-operating trailing axes.

    Parameters
    ----------
    A : ndarray of shape (M, N, M, N, ...)
        The left-hand side array of the equation system.
    b : ndarray of shape (M, N, ...)
        The right-hand side array of the equation system.
    n : int, optional
        The dimension of the unknowns, must be greater or equal zero (default is 1).
    solve : callable, optional
        A function with a compatible signature to :func:`numpy.linalg.solve`
        (default is :func:`numpy.linalg.solve`).
    **kwargs : dict, optional
        Optional keyword arguments are passed to the callable ``solve(A, b, **kwargs)``.

    Returns
    -------
    x : ndarray of shape (M, N, ...)
        The solved array of unknowns.

    Notes
    -----
    The first two axes of the rhs ``b`` and the first four axes of the lhs ``A`` are the
    tensor dimensions and all remaining trailing axes are treated as batch dimensions.
    This function finds :math:`x_{kl}` for Eq. :eq:`solve-system-nd`.

    ..  math::
        :label: solve-system-nd

        A_{i...j...} : x_{j...} = b_{i...}

    Examples
    --------
    >>> import numpy as np
    >>> import felupe as fem
    >>>
    >>> np.random.seed(855436)
    >>>
    >>> A = np.random.rand(3, 3, 3, 3, 3, 3, 2, 4)
    >>> b = np.ones((3, 3, 3, 1, 1))
    >>>
    >>> x = fem.math.solve_nd(A, b, n=3)

    >>> x[..., 0, 0]
    array([[[ 0.63499407, -1.32830401, -1.14593   ],
            [ 1.38045295, -1.03504171,  0.02272754],
            [-0.40106257,  1.6447736 ,  1.87685462]],

           [[ 0.22602567,  0.53675077,  0.37876301],
            [-0.5807506 ,  1.20191896,  0.46334835],
            [-0.90352229,  0.45065787, -0.53026653]],

           [[ 0.3551602 , -1.06411506,  0.01788346],
            [-1.33151311,  1.76817965,  0.00544784],
            [ 0.67065754,  0.43405431, -1.67714269]]])

    >>> x.shape
    (3, 3, 3, 2, 4)

    """

    # broadcast matrix-axes of lhs
    Ashape_new = list(A.shape)
    Ashape_new[:n] = Ashape_new[n : 2 * n] = np.broadcast_shapes(
        A.shape[:n], A.shape[n : 2 * n]
    )
    A = np.broadcast_to(A, Ashape_new)

    # broadcast matrix-axes of rhs
    bshape_new = list(b.shape)
    bshape_new[:n] = np.broadcast_shapes(A.shape[:n], b.shape[:n])
    b = np.broadcast_to(b, bshape_new)

    # trailing (batch) axes of broadcasted output
    shape = b.shape[:n]
    size = np.prod(shape, dtype=int)
    trax = np.broadcast_shapes(b.shape[n:], A.shape[2 * n :])

    # flatten and reshape A to a 2d-matrix of shape (..., M * N * ..., M * N * ...) and
    # b to a 1d-vector of shape (..., M * N * ..., 1) with batches at leading axis
    b_1d = np.einsum("ik...->...ik", b.reshape(size, 1, -1))
    A_1d = np.einsum("ij...->...ij", A.reshape(size, size, -1))

    # move the batch-dimensions to the back and reshape x
    return np.einsum("...ik->ik...", solve(A_1d, b_1d, **kwargs)).reshape(*shape, *trax)


def solve_2d(A, b, solve=np.linalg.solve, **kwargs):
    r"""Solve a linear equation system for two-dimensional unknowns with optional
    elementwise-operating trailing axes.

    Parameters
    ----------
    A : ndarray of shape (M, N, M, N, ...)
        The left-hand side array of the equation system.
    b : ndarray of shape (M, N, ...)
        The right-hand side array of the equation system.
    n : int, optional
        The dimension of the unknowns (default is 1).
    solve : callable, optional
        A function with a compatible signature to :func:`numpy.linalg.solve`
        (default is :func:`numpy.linalg.solve`).
    **kwargs : dict, optional
        Optional keyword arguments are passed to the callable ``solve(A, b, **kwargs)``.

    Returns
    -------
    x : ndarray of shape (M, N, ...)
        The solved array of unknowns.

    Notes
    -----
    The first two axes of the rhs ``b`` and the first four axes of the lhs ``A`` are the
    tensor dimensions and all remaining trailing axes are treated as batch dimensions.
    This function finds :math:`x_{kl}` for Eq. :eq:`solve-system-2d`.

    ..  math::
        :label: solve-system-2d

        A_{ijkl} : x_{kl} = b_{ij}

    Examples
    --------
    >>> import numpy as np
    >>> import felupe as fem
    >>>
    >>> np.random.seed(855436)
    >>>
    >>> A = np.random.rand(3, 3, 3, 3, 2, 4)
    >>> b = np.ones((3, 3, 2, 4))
    >>>
    >>> x = fem.math.solve_2d(A, b)

    >>> x[..., 0, 0]
    array([[ 1.1442917 ,  2.14516919,  2.00237954],
           [-0.69463749, -2.46685827, -7.21630899],
           [ 4.44825615,  1.35899745,  1.08645703]])

    >>> x.shape
    (3, 3, 2, 4)

    """
    return solve_nd(A=A, b=b, solve=solve, n=2, **kwargs)



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import numpy as np

import felupe as fem


def test_visco_newton():
    import tensortrax as tr
    from tensortrax.math import trace
    from tensortrax.math.linalg import det, inv
    from tensortrax.math.special import from_triu_1d, triu_1d

    import felupe.constitution.tensortrax as mat

    mesh = fem.Rectangle(n=3)
    region = fem.RegionQuad(mesh)
    field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])
    boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)

    @mat.isochoric_volumetric_split
    def finite_strain_viscoelastic_newton(C, Cin, mu, eta, dtime):
        "Finite Strain Viscoelastic material formulation."

        def evolution(Ci, Cin, C, mu, eta, dtime):
            "Viscoelastic evolution equation."
            return mu / eta * C - (Ci - Cin) / dtime

        # update of state variables by evolution equation
        Cin = from_triu_1d(Cin, like=C)
        Ci = fem.newtonrhapson(
            x0=Cin,
            fun=tr.function(evolution, ntrax=C.ntrax),
            jac=tr.jacobian(evolution, ntrax=C.ntrax),
            solve=fem.math.solve_2d,
            args=(Cin, C.x, mu, eta, dtime),
            verbose=0,
        ).x
        Ci *= det(Ci) ** (-1 / 3)

        # first invariant of elastic part of right Cauchy-Green deformation tensor
        I1 = trace(C @ inv(Ci))

        # strain energy function and state variable
        return mu / 2 * (I1 - 3), triu_1d(Ci)

    umat = mat.Hyperelastic(
        finite_strain_viscoelastic_newton, mu=1.0, eta=1.0, dtime=0.1, nstatevars=6
    )
    solid = fem.SolidBodyNearlyIncompressible(umat, field, bulk=5000)

    move = fem.math.linsteps([0, 0.3], num=3)
    step = fem.Step(
        items=[solid], ramp={boundaries["move"]: move}, boundaries=boundaries
    )
    job = fem.CharacteristicCurve(steps=[step], boundary=boundaries["move"]).evaluate()
    assert np.all([norm[-1] < 1e-7 for norm in job.fnorms])


def test_visco_newton_jax():
    import jax
    from jax.numpy import trace
    from jax.numpy.linalg import det, inv
    from jax.scipy.optimize import minimize

    import felupe.constitution.jax as mat

    mesh = fem.Rectangle(n=3)
    region = fem.RegionQuad(mesh)
    field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])
    boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)

    @mat.isochoric_volumetric_split
    def finite_strain_viscoelastic_newton(C, statevars, mu, eta, dtime):
        "Finite Strain Viscoelastic material formulation."

        def evolution(Ci, Cin, C, mu, eta, dtime):
            "Viscoelastic evolution equation."
            Ci = Ci.reshape(3, 3)
            residuals = mu / eta * C - (Ci - Cin) / dtime
            return residuals.ravel() @ residuals.ravel()

        # update of state variables by evolution equation
        Cin = statevars[:9].reshape(3, 3)
        Ci = minimize(
            evolution,
            x0=Cin.ravel(),
            args=(Cin, jax.lax.stop_gradient(C), mu, eta, dtime),
            method="BFGS",
        ).x.reshape(3, 3)
        Ci *= det(Ci) ** (-1 / 3)

        # first invariant of elastic part of right Cauchy-Green deformation tensor
        I1 = trace(C @ inv(Ci))

        # strain energy function and state variable
        statevars_new = Ci.ravel()
        return mu / 2 * (I1 - 3), statevars_new

    umat = mat.Hyperelastic(
        finite_strain_viscoelastic_newton,
        mu=1.0,
        eta=1.0,
        dtime=0.1,
        nstatevars=9,
        jit=True,
    )
    solid = fem.SolidBodyNearlyIncompressible(umat, field, bulk=5000)

    move = fem.math.linsteps([0, 0.3], num=3)
    step = fem.Step(
        items=[solid], ramp={boundaries["move"]: move}, boundaries=boundaries
    )
    job = fem.CharacteristicCurve(steps=[step], boundary=boundaries["move"])
    job.evaluate(tol=1e-3)

    assert np.all([fnorm[-1] < 1e-3 for fnorm in job.fnorms])


if __name__ == "__main__":
    test_visco_newton()
    test_visco_newton_jax()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._base import Element


class Vertex(Element):
    r"""A vertex element formulation with constant shape functions.

    Notes
    -----
    The vertex element is defined by one point.
    """

    def __init__(self):
        self.points = np.array([[0.0]])
        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "vertex"

    def function(self, r):
        "Return the shape functions at given coordinate (r)."
        return np.ones(1)

    def gradient(self, r):
        "Return the gradient of shape functions at given coordinate (r)."
        return np.zeros((1, 1))

    def hessian(self, rs):
        "Return the hessian of shape functions at given coordinate (r)."
        return np.zeros((1, 1, 1))
